import unittest
import tempfile
import os
from pathlib import Path

# Load release module as in other tests
from gway import gw


class ChangelogUpdateTests(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.TemporaryDirectory()
        self.base = Path(self.tmp.name)
        self.old_cwd = Path.cwd()
        os.chdir(self.base)

    def tearDown(self):
        os.chdir(self.old_cwd)
        self.tmp.cleanup()

    def _write_changelog(self, text: str):
        Path("CHANGELOG.rst").write_text(text)

    def test_update_changelog_underline_matches_header(self):
        content = """Changelog
=========

Unreleased
----------
- first change
"""
        self._write_changelog(content)
        gw.release.update_changelog("1.2.3", "abcdef")
        text = Path("CHANGELOG.rst").read_text()
        lines = text.splitlines()
        header = "1.2.3 [build abcdef]"
        idx = lines.index(header)
        underline = lines[idx + 1]
        self.assertEqual(len(underline), len(header))

if __name__ == "__main__":
    unittest.main()



# projects/clock.py

import re
import time
import requests
import datetime

#       If there is missing functionality we need for other projects, add it.

def now(*, utc=False) -> "datetime":
    """Return the current datetime object."""
    return datetime.datetime.now(datetime.timezone.utc) if utc else datetime.datetime.now()


def plus(*, seconds=0, utc=False) -> "datetime":
    """Return current datetime plus given seconds."""
    base = now(utc=utc)
    return base + datetime.timedelta(seconds=seconds)


def minus(*, seconds=0, utc=False) -> "datetime":
    """Return current datetime plus given seconds."""
    base = now(utc=utc)
    return base - datetime.timedelta(seconds=seconds)


def timestamp(*, utc=False) -> str:
    """Return the current timestamp in ISO-8601 format."""
    return now(utc=utc).isoformat().replace("+00:00", "Z" if utc else "")


...


def to_download(filesize):
    """ 
    Prompt: Create a python function that takes a file size such as 100 MB or 1.76 GB 
    (pick a wide array of units) and then calculates the possible time to download 
    it within 4 ranges. You choose the ranges logarithmically. Then, perform a quick 
    check against google to let the user know what their current range is.
    """

    # 1. Size parsing
    def parse_size(size_str):
        """
        Parse a size string like '1.76 GB', '500kb', '1024 B', '3.2 GiB'
        into a number of bytes (float).
        Accepts decimal (k=1e3) or binary (Ki=2**10) prefixes.
        """
        size_str = size_str.strip()

        # Handle plain bytes (e.g. '123 B')
        plain_b_match = re.match(r"^([\d\.]+)\s*(bytes?|b)$", size_str, re.IGNORECASE)
        if plain_b_match:
            return float(plain_b_match.group(1))

        # Handle KB/MB/etc
        pattern = r"^([\d\.]+)\s*([kKmMgGtTpP])([iI])?[bB]?$"
        m = re.match(pattern, size_str)
        if not m:
            raise ValueError(f"Unrecognized size format: {size_str!r}")
        num, prefix, binary = m.group(1, 2, 3)
        num = float(num)
        exp = {"k": 1, "m": 2, "g": 3, "t": 4, "p": 5}[prefix.lower()]
        if binary:
            factor = 2 ** (10 * exp)
        else:
            factor = 10 ** (3 * exp)
        return num * factor

    # 2. Human-friendly duration
    def format_duration(seconds):
        if seconds < 1:
            return f"{seconds*1000:.1f} ms"
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        parts = []
        if h: parts.append(f"{int(h)} h")
        if m: parts.append(f"{int(m)} m")
        if s or not parts: parts.append(f"{s:.1f} s")
        return " ".join(parts)

    # 3. Estimate times at four speeds
    SPEED_BRACKETS = [
        (1e3,   "1 kB/s (≈8 kbps)"),
        (1e5,   "100 kB/s (≈0.8 Mbps)"),
        (1e7,   "10 MB/s (≈80 Mbps)"),
        (1e9,   "1 GB/s (≈8 Gbps)"),
    ]

    def estimate_download_times(size_str):
        size_bytes = parse_size(size_str)
        return [(label, format_duration(size_bytes / speed))
                for speed, label in SPEED_BRACKETS]

    # 4. Quick live check against Google
    def measure_current_speed(test_url=None):
        if test_url is None:
            test_url = ("https://www.google.com/images/branding/"
                        "googlelogo/2x/googlelogo_color_272x92dp.png")
        start = time.time()
        r = requests.get(test_url, stream=True, timeout=10)
        total = 0
        for chunk in r.iter_content(1024):
            total += len(chunk)
        elapsed = time.time() - start
        return total / elapsed  # bytes per second

    def classify_speed(bps):
        for i, (speed, label) in enumerate(SPEED_BRACKETS):
            # slower than midpoint to next bracket → this bracket
            if i == len(SPEED_BRACKETS)-1 or \
               bps < speed * ((SPEED_BRACKETS[i+1][0]/speed)**0.5):
                return label
        return SPEED_BRACKETS[-1][1]

    # 5. Putting it all together
    def download_time_report(size_str):
        print(f"Estimating download times for {size_str!r}:\n")
        for label, t in estimate_download_times(size_str):
            print(f" • At {label}: {t}")
        try:
            print("\nMeasuring your current download speed…")
            speed = measure_current_speed()
            # human = time to download 1 B
            human = format_duration(parse_size('1 B') / speed)
            print(f" → Detected ≈{speed/1024:.1f}\u00A0kB/s ({human})")
            bracket = classify_speed(speed)
            print(f"   You’re in the **{bracket}** range.")
        except Exception as e:
            print(f" (Could not measure live speed: {e})")
            raise

    download_time_report(filesize)



# file: projects/odoo.py

from xmlrpc import client
from datetime import datetime, timedelta
from fastapi import FastAPI, Request, Form
from fastapi.responses import HTMLResponse, RedirectResponse
from datetime import datetime
from contextlib import asynccontextmanager
import asyncio
from gway import gw


def execute_kw(*args, model: str, method: str, **kwargs) -> dict:
    """
    A generic function to directly interface with Odoo's execute_kw method.

    Parameters:
        model (str): The Odoo model to interact with (e.g., 'sale.order').
        method (str): The method to call on the model (e.g., 'read', 'write').
        args (list): Positional arguments to pass to the method.
        kwargs (dict): Keyword arguments to pass to the method.

    Returns:
        dict: The result of the execute_kw call.
    """
    url = gw.resolve("[ODOO_BASE_URL]")
    db_name = gw.resolve("[ODOO_DB_NAME]")
    username = gw.resolve("[ODOO_ADMIN_USER]")
    password = gw.resolve("[ODOO_ADMIN_PASSWORD]")

    gw.info(f"Odoo Execute: {model=} {method=} @ {url=} {db_name=} {username=}")
    if url.startswith("[") or "ODOO_BASE_URL" in url:
        gw.halt("Odoo XML-RPC url not configured. Please set ODOO_BASE_URL correctly.")
    try:
        common_client = client.ServerProxy(f"{url}/xmlrpc/2/common")
    except Exception as e:
        gw.exception(f"Error with ServerProxy setup", e)
        raise
    gw.debug(f"ServerProxy client: {common_client}")
    try:
        uid = common_client.authenticate(db_name, username, password, {})
    except Exception as e:
        gw.error(f"Error with Odoo authentication: {e}")
        print(f"( Did you forget to specify the correct --client? )")
        raise

    try:
        models_client = client.ServerProxy(f"{url}/xmlrpc/2/object")
        for reserved in ("db_name", "uid", "password", "model", "method"):
            gw.warning(f"Removing reserved keyword: {reserved}")
            kwargs.pop(reserved, None)
        gw.debug(f"Model client call execute_kw {model}.{method} with {args=} {kwargs=}")
        result = models_client.execute_kw(db_name, uid, password, model, method, *args, **kwargs)
        return result
    except Exception as e:
        gw.error(f"Error executing {model}.{method}: {e}")
        raise


def fetch_quotes(
    *,
    state='draft',
    older_than=None,
    salesperson=None,
    customer=None,
    **kwargs
):
    """
    Fetch quotes/quotations from Odoo with optional filters.

    Parameters:
        state (str): Filter quotations by their state. Default is 'draft'.
        older_than (int, optional): Filter quotations older than a specific number of days.
        salesperson (str, optional): Filter quotations by the salesperson's name or part of it.
        customer (str, optional): Filter quotations by the customer's name or part of it.
        kwargs (list, optional): Additional domain filters for the query.

    Returns:
        dict: The fetched quotations.
    """
    model = 'sale.order'
    method = 'search_read'

    domain_filter = [('state', '=', state)]
    if older_than:
        cutoff_date = (datetime.now() - timedelta(days=older_than)).strftime('%Y-%m-%d')
        domain_filter.append(('create_date', '<=', cutoff_date))
    if salesperson:
        domain_filter.append(('user_id.name', 'ilike', salesperson))
    if customer:
        domain_filter.append(('partner_id.name', 'ilike', customer))
    if kwargs:
        domain_filter.extend(kwargs)
    fields_to_fetch = ['name', 'amount_total', 'create_date', 'user_id', 'partner_id']
    try:
        result = execute_kw(
            [domain_filter], {'fields': fields_to_fetch},
            model=model, method=method
        )
        return result
    except Exception as e:
        gw.error(f"Error fetching quotations: {e}")
        raise


def fetch_products(*, name=None, latest_quotes=None):
    """
    Fetch the list of non-archived products from Odoo.
    If a name is provided, use it as a partial filter on the product name.
    """
    model = 'product.product'
    method = 'search_read'
    domain_filter = [('active', '=', True)]  # Non-archived products have active=True
    if name:
        domain_filter.append(('name', 'ilike', name))
    
    fields_to_fetch = ['name', 'list_price']  # Add fields as needed
    result = execute_kw(
        [domain_filter], {'fields': fields_to_fetch},
        model=model, method=method
    )
    return result


def fetch_customers(
    *,
    name=None,
    email=None,
    phone=None,
    country=None,
    latest_quotes=None,
    **kwargs
):
    """
    Fetch customers from Odoo with optional filters.

    Parameters:
        name (str, optional): Filter customers by their name or part of it.
        email (str, optional): Filter customers by their email address or part of it.
        phone (str, optional): Filter customers by their phone number or part of it.
        country (str, optional): Filter customers by their country name or part of it.
        **kwargs: Additional filters to be applied, passed as key-value pairs.

    Returns:
        dict: The fetched customers.
    """

    model = 'res.partner'
    method = 'search_read'

    # Start with an empty domain filter
    domain_filter = []

    if name:
        domain_filter.append(('name', 'ilike', name))
    if email:
        domain_filter.append(('email', 'ilike', email))
    if phone:
        domain_filter.append(('phone', 'ilike', phone))
    if country:
        domain_filter.append(('country_id.name', 'ilike', country))
    for field, value in kwargs.items():
        domain_filter.append((field, 'ilike', value))

    fields_to_fetch = ['name', 'create_date']
    try:
        result = execute_kw(
            [domain_filter], {'fields': fields_to_fetch},
            model=model, method=method
        )
        return result
    except Exception as e:
        gw.error(f"Error fetching customers: {e}")
        raise


def fetch_order(order_id):
    """
    Fetch the details of a specific order by its ID from Odoo, including all line details.
    """
    order_model = 'sale.order'
    order_method = 'read'
    line_model = 'sale.order.line'
    line_method = 'search_read'
    
    order_fields = ['name', 'amount_total', 'partner_id', 'state']
    line_fields = ['product_id', 'name', 'price_unit', 'product_uom_qty']
    
    # Check if order_id is a string that starts with 'S' and fetch by name instead of ID
    if isinstance(order_id, str) and order_id.startswith('S'):
        order_domain_filter = [('name', '=', order_id)]
        order_result = execute_kw(
            order_model, 'search_read', [order_domain_filter], {'fields': order_fields})
        if order_result:
            order_id = order_result[0]['id']
        else:
            return {'error': 'Order not found.'}
    else:
        order_result = execute_kw(
            [[order_id]], {'fields': order_fields},
            model=order_model, method=order_method,
        )

    line_domain_filter = [('order_id', '=', order_id)]
    line_result = execute_kw(
        [line_domain_filter], {'fields': line_fields},
        model=line_model, method=line_method,
    )
    
    result = {
        'order_info': order_result,
        'line_details': line_result
    }
    
    return result
        

def fetch_templates(*, name=None, active=True, **kwargs):
    """
    Fetch available quotation templates from Odoo with optional filters.

    Parameters:
        name (str, optional): Filter templates by name or part of it.
        active (bool): Whether to include only active templates. Defaults to True.
        **kwargs: Additional filters as key-value pairs.

    Returns:
        dict: The fetched quotation templates.
    """
    model = 'sale.order.template'
    method = 'search_read'
    
    domain_filter = []
    if name:
        domain_filter.append(('name', 'ilike', name))
    if active is not None:
        domain_filter.append(('active', '=', active))
    for field, value in kwargs.items():
        domain_filter.append((field, '=', value))

    fields_to_fetch = ['name', 'number_of_days', 'active']
    
    try:
        result = execute_kw(
            [domain_filter], {'fields': fields_to_fetch},
            model=model, method=method
        )
        return result
    except Exception as e:
        gw.error(f"Error fetching quotation templates: {e}")
        raise


def create_quote(*, customer, template_id, validity=None, notes=None):
    """
    Create a new quotation using a specified template and customer name.

    Parameters:
        customer (str): The name (or partial name) of the customer to link to the quote.
        template_id (int): The ID of the quotation template to use.
        validity (str, optional): The expiration date for the quote in 'YYYY-MM-DD' format.
        notes (str, optional): Internal notes or message to include in the quote.

    Returns:
        dict: The created quotation details.
    """
    # Step 1: Lookup the customer ID
    customer_result = fetch_customers(name=customer)
    if not customer_result:
        return {'error': f"No customer found matching name: {customer}"}
    
    customer_id = customer_result[0]['id']

    # Step 2: Create the quote using the template
    model = 'sale.order'
    method = 'create'

    values = {
        'partner_id': customer_id,
        'sale_order_template_id': template_id,
    }

    if validity:
        values['validity_date'] = validity
    if notes:
        values['note'] = notes

    try:
        quote_id = execute_kw(
            [values], {},
            model=model, method=method
        )
    except Exception as e:
        gw.error(f"Error creating quote: {e}")
        raise

    # Step 3: Return full quote details
    return fetch_order(quote_id)


def send_chat(message: str, *, username: str = "[ODOO_USERNAME]") -> bool:
    """
    Send a chat message to an Odoo user by username.
    """
    user_info = get_user_info(username=username)
    if not user_info:
        return False

    user_id = user_info["id"]
    return execute_kw(
        model="mail.channel",
        method="message_post",
        kwargs={
            "partner_ids": [(4, user_id)],
            "body": message,
            "message_type": "comment",
            "subtype_xmlid": "mail.mt_comment",
        },
    )


def read_chat(*, 
        unread: bool = True, 
        username: str = "[ODOO_USERNAME]", 
    ) -> list[dict]:
    """
    Read chat messages from an Odoo user by username.
    If unread is True, only return unread messages.
    """
    username = gw.resolve(username) if isinstance(username, str) else username
    user_info = get_user_info(username=username)
    if not user_info: return []

    user_id = user_info["id"]
    domain = [["author_id", "=", user_id]]
    if unread:
        domain.append(["message_read", "=", False])

    messages = execute_kw(
        model="mail.message",
        method="search_read",
        domain=domain,
        fields=["id", "body", "date", "author_id", "message_type"],
    )
    return messages


def get_user_info(*, username: str) -> dict:
    """Retrieve Odoo user information by username."""
    user_data = execute_kw(
        model="res.users",
        method="search_read",
        domain=[["login", "=", username]],
        fields=["id", "name", "login"],
    )
    if not user_data:
        gw.error(f"User not found: {username}")
        return None
    return user_data[0]  # Return the first (and likely only) match.


chatbot_log: list[dict] = []

def setup_chatbot_app(*, 
            path="/chatbot", username="[ODOO_USERNAME]", alias="Operator", apps=None
        ):
    """
    Create a FastAPI app (or append to existing ones) serving a chatbot UI and logic.
    """

    last_seen = {}
    username = gw.resolve(username) if isinstance(username, str) else username

    def log_msg(direction, msg):
        chatbot_log.append({
            "timestamp": datetime.utcnow().isoformat(),
            "direction": direction,
            "message": msg,
        })

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        async def poll_messages():
            while True:
                try:
                    messages = read_chat(username=username)
                    for msg in messages:
                        msg_id = msg["id"]
                        if msg_id not in last_seen:
                            log_msg("in", msg["body"])
                            last_seen[msg_id] = True
                    await asyncio.sleep(5)
                except Exception as e:
                    gw.error(f"Chatbot polling error: {e}")
                    await asyncio.sleep(10)
        asyncio.create_task(poll_messages())
        yield

    app = FastAPI(lifespan=lifespan)

    @app.get(path, response_class=HTMLResponse)
    async def ui(request: Request):
        messages = read_chat(username=username, unread=False)
        html = f"""
        <html>
        <head>
            <title>Chatbot Interface</title>
            <style>
                body {{ font-family: sans-serif; padding: 1rem; }}
                .log {{ background: #f9f9f9; padding: 1em; margin-bottom: 1em; border: 1px solid #ccc; }}
                .incoming {{ color: darkblue; }}
                .outgoing {{ color: darkgreen; }}
                form {{ margin-top: 1em; }}
            </style>
        </head>
        <body>
            <h1>Chatbot: {username}</h1>
            <div class="log">
        """
        for msg in reversed(messages[-20:]):
            cls = "incoming" if msg["author_id"][1] != alias else "outgoing"
            html += f"<div class='{cls}'><b>{msg['author_id'][1]}</b>: {msg['body']}</div>"
        html += f"""
            </div>
            <form method="post">
                <label>Alias: <input name="alias" value="{alias}" /></label><br />
                <textarea name="body" rows="3" cols="60" placeholder="Type response..."></textarea><br/>
                <button type="submit">Send</button>
            </form>
        </body>
        </html>
        """
        return HTMLResponse(content=html, status_code=200)

    @app.post(path)
    async def post_message(alias: str = Form(...), body: str = Form(...)):
        if body.strip():
            send_chat(body, username=username)
            log_msg("out", body)
        return RedirectResponse(url=path, status_code=303)

    if apps is None:
        return [app]
    elif isinstance(apps, list):
        return apps + [app]
    else:
        return [apps, app]

# projects/odoo.py

def find_quotes(
    *,
    product,
    quantity: int = 1,
    state: str = 'draft',
    **kwargs
):
    """
    Find all sale quotes that contain a given product (by id or name substring) with at least the given quantity.

    Parameters:
        product (str or int): Product ID or partial name.
        quantity (int): Minimum quantity of the product in the quote. Default is 1.
        state (str): Odoo sale order state (default: 'draft' for quotations).
        **kwargs: Additional domain filters for sale.order.

    Returns:
        list: List of matching sale orders (quotes) with product line details.
    """
    gw.info(f"Finding quotes for {product=} {quantity=}")
    # Step 1: Resolve product id if necessary
    product_id = None

    # Try converting product to integer (for id)
    try:
        product_id = int(product)
        product_name = None
    except (ValueError, TypeError):
        # Search by product name substring
        results = fetch_products(name=product)
        if not results:
            return {"error": f"No products found matching: {product}"}
        if len(results) > 1:
            return {
                "error": f"Ambiguous product name '{product}', matches: " +
                         ", ".join([f"{p['id']}: {p['name']}" for p in results])
            }
        product_id = results[0]['id']
        product_name = results[0]['name']
        gw.info(f"Resolved product '{product}' to id {product_id} ('{product_name}')")
    
    # Step 2: Find sale order lines matching product + min quantity
    line_model = 'sale.order.line'
    line_method = 'search_read'
    domain_lines = [
        ('product_id', '=', product_id),
        ('product_uom_qty', '>=', quantity)
    ]
    line_fields = ['order_id', 'product_id', 'product_uom_qty', 'name']
    sale_lines = execute_kw(
        [domain_lines],
        {'fields': line_fields},
        model=line_model,
        method=line_method
    )
    if not sale_lines:
        return {"result": [], "info": f"No quotes found with product {product_id} and quantity >= {quantity}"}
    
    # Step 3: Collect all order_ids found in lines
    order_ids = list(set(l['order_id'][0] if isinstance(l['order_id'], (list, tuple)) else l['order_id'] for l in sale_lines))
    if not order_ids:
        return {"result": [], "info": f"No matching quotes found."}
    
    # Step 4: Fetch quotes for those order_ids with optional state filter
    order_model = 'sale.order'
    order_method = 'search_read'
    domain_orders = [('id', 'in', order_ids)]
    if state:
        domain_orders.append(('state', '=', state))
    # Add any extra filters from kwargs
    for key, value in kwargs.items():
        domain_orders.append((key, '=', value))
    fields_to_fetch = ['name', 'amount_total', 'create_date', 'user_id', 'partner_id', 'state']

    quotes = execute_kw(
        [domain_orders],
        {'fields': fields_to_fetch},
        model=order_model,
        method=order_method
    )

    # Step 5: Attach relevant line(s) for each quote
    quote_lines_by_order = {}
    for line in sale_lines:
        oid = line['order_id'][0] if isinstance(line['order_id'], (list, tuple)) else line['order_id']
        quote_lines_by_order.setdefault(oid, []).append({
            "product_id": line['product_id'],
            "qty": line['product_uom_qty'],
            "line_name": line['name'],
        })
    # Attach to each quote
    for quote in quotes:
        quote['matching_lines'] = quote_lines_by_order.get(quote['id'], [])

    return quotes


def fetch_projects(*, name=None):
    """Fetch projects by partial name."""
    model = 'project.project'
    method = 'search_read'

    domain_filter = []
    if name:
        domain_filter.append(('name', 'ilike', name))

    fields_to_fetch = ['name']
    result = execute_kw(
        [domain_filter], {'fields': fields_to_fetch},
        model=model, method=method
    )
    return result


def create_task(
    *,
    title: str | None = None,
    project: str | int = '[ODOO_DEFAULT_PROJECT]',
    customer: str | None = None,
    phone: str | None = None,
    notes: str | None = None,
    new_customer: bool = False,
):
    """Create an Odoo project task with optional customer creation.

    If ``title`` is not provided but ``customer`` is, the task title defaults
    to the customer name.
    """

    # Resolve default project value from environment
    if isinstance(project, str):
        project_resolved = gw.resolve(project)
    else:
        project_resolved = project

    # Determine project_id
    try:
        project_id = int(project_resolved)
    except (ValueError, TypeError):
        projects = fetch_projects(name=project_resolved)
        if not projects:
            return {'error': f"Project not found: {project_resolved}"}
        if len(projects) > 1:
            gw.warning(f"Multiple projects match '{project_resolved}', using first")
        project_id = projects[0]['id']

    # Handle customer lookup / creation
    customer_id = None
    if customer:
        if new_customer:
            values = {'name': customer}
            if phone:
                values['phone'] = phone
            if notes:
                values['comment'] = notes
            customer_id = execute_kw(
                [values], {},
                model='res.partner', method='create'
            )
        else:
            result = fetch_customers(name=customer)
            if not result:
                return {'error': f"Customer not found: {customer}"}
            customer_id = result[0]['id']

    if title is None:
        if customer:
            title = customer
        else:
            return {'error': 'title or customer required'}

    description_parts = []
    if phone:
        description_parts.append(f"Phone: {phone}")
    if notes:
        description_parts.append(notes)
    description = '\n'.join(description_parts)

    task_vals = {
        'name': title,
        'project_id': project_id,
    }
    if customer_id:
        task_vals['partner_id'] = customer_id
    if description:
        task_vals['description'] = description

    task_id = execute_kw(
        [task_vals], {},
        model='project.task', method='create'
    )
    task = execute_kw(
        [[task_id]], {'fields': ['id', 'name', 'project_id', 'partner_id', 'description']},
        model='project.task', method='read'
    )
    return task



# file: projects/web/error.py

from gway import gw

def view_debug_error(
    *,
    title="GWAY Debug Error",
    message="An error occurred.",
    err=None,
    status=500,
    default=None,
    extra=None
):
    """
    Render a debug error view with detailed traceback and request info.
    """
    from bottle import request, response
    import traceback
    import html

    tb_str = ""
    if err:
        tb_str = "".join(traceback.format_exception(type(err), err, getattr(err, "__traceback__", None)))

    extra = extra or ""
    debug_content = f"""
    <html>
    <head>
        <title>{html.escape(title)}</title>
        <style>
            body {{ font-family: monospace, sans-serif; background: #23272e; color: #e6e6e6; }}
            .traceback {{ background: #16181c; color: #ff8888; padding: 1em; border-radius: 5px; margin: 1em 0; white-space: pre; }}
            .kv {{ color: #6ee7b7; }}
            .section {{ margin-bottom: 2em; }}
            h1 {{ color: #ffa14a; }}
            a {{ color: #69f; }}
            .copy-btn {{ margin: 1em 0; background:#333;color:#fff;padding:0.4em 0.8em;border-radius:4px;cursor:pointer;border:1px solid #aaa; }}
        </style>
    </head>
    <body>
        <h1>{html.escape(title)}</h1>
        <div id="debug-content">
            <div class="section"><b>Message:</b> {html.escape(str(message) or "")}</div>
            <div class="section"><b>Error:</b> {html.escape(str(err) or "")}</div>
            <div class="section"><b>Path:</b> {html.escape(request.path or "")}<br>
                                 <b>Method:</b> {html.escape(request.method or "")}<br>
                                 <b>Full URL:</b> {html.escape(request.url or "")}</div>
            <div class="section"><b>Query:</b> {html.escape(str(dict(request.query)) or "")}</div>
            <div class="section"><b>Form:</b> {html.escape(str(getattr(request, "forms", "")) or "")}</div>
            <div class="section"><b>Headers:</b> {html.escape(str(dict(request.headers)) or "")}</div>
            <div class="section"><b>Cookies:</b> {html.escape(str(dict(request.cookies)) or "")}</div>
            <div class="section"><b>Traceback:</b>
                <div class="traceback">{html.escape(tb_str or '(no traceback)')}</div>
            </div>
        </div>
        {extra}
        <div><a href="{html.escape(default or gw.web.app.default_home())}">&#8592; Back to home</a></div>
    </body>
    </html>
    """
    response.status = status
    response.content_type = "text/html"
    return debug_content

def redirect(message="", *, err=None, default=None, view_name=None):
    """
    GWAY error/redirect handler.
    Deprecated: 'view_name'. Now uses gw.web.app.current_endpoint.
    """
    from bottle import request, response

    debug_enabled = bool(getattr(gw, "debug", False))
    visited = gw.web.cookies.get("visited", "")
    visited_items = visited.split("|") if visited else []

    # --- DEPRECATED: view_name, use gw.web.app.current_endpoint instead ---
    if view_name is not None:
        import warnings
        warnings.warn(
            "redirect(): 'view_name' is deprecated. Use gw.web.app.current_endpoint instead.",
            DeprecationWarning
        )
    curr_view = getattr(gw.web.app, "current_endpoint", None)
    view_key = curr_view() if callable(curr_view) else curr_view
    if not view_key and view_name:
        view_key = view_name

    pruned = False
    if view_key and gw.web.cookies.accepted():
        norm_broken = (view_key or "").replace("-", " ").replace("_", " ").title().lower()
        new_items = []
        for v in visited_items:
            title = v.split("=", 1)[0].strip().lower()
            if title == norm_broken:
                pruned = True
                continue
            new_items.append(v)
        if pruned:
            gw.web.cookies.set("visited", "|".join(new_items))
            visited_items = new_items

    if debug_enabled:
        return view_debug_error(
            title="GWAY Debug Error",
            message=message,
            err=err,
            status=500,
            default=default
        )

    response.status = 302
    response.set_header("Location", default or gw.web.app.default_home())
    return ""

def unauthorized(message="Unauthorized: You do not have access to this resource.", *, err=None, default=None):
    """
    If in debug mode: show detailed error.
    If not in debug: return a 401 Unauthorized and a WWW-Authenticate header to trigger the browser auth dialog.
    """
    from bottle import response, request

    debug_enabled = bool(getattr(gw, "debug", False))
    if debug_enabled:
        from bottle import request
        import html
        orig_link = f'<div><a href="{html.escape(request.url)}">Go to original page</a></div>'
        return view_debug_error(
            title="401 Unauthorized",
            message=message,
            err=err,
            status=401,
            default=default,
            extra=orig_link
        )

    # 401 with auth header = browser will prompt for password
    response.status = 401
    response.headers['WWW-Authenticate'] = 'Basic realm="GWAY"'
    response.content_type = "text/html"
    import html
    target = request.headers.get("Referer") or default or gw.web.app.default_home()
    esc_target = html.escape(target)
    esc_msg = html.escape(str(message))
    html_page = f"""
    <html>
    <head>
        <meta http-equiv='refresh' content='5; url={esc_target}'>
        <title>401 Unauthorized</title>
    </head>
    <body>
        <h1>401 Unauthorized</h1>
        <p>{esc_msg}</p>
        <p>Username and password are required.</p>
        <p>You will be redirected in 5 seconds. If not, <a href='{esc_target}'>click here</a>.</p>
    </body>
    </html>
    """
    return html_page



# file: projects/cli.py
"""CLI helper utilities.

This project provides functions related to the command
line interface. The ``completions`` function is a prototype
to inspect available project commands for custom shell
completion scripts.
"""

from __future__ import annotations

import inspect
from gway import gw
from gway.structs import Project


def _walk(ns: Project, parts: list[str]) -> list[list[str]]:
    commands: list[list[str]] = []
    for name in dir(ns):
        if name.startswith("_"):
            continue
        obj = getattr(ns, name)
        if isinstance(obj, Project):
            commands.extend(_walk(obj, parts + [name]))
        elif inspect.isfunction(obj):
            commands.append(parts + [name])
    return commands


def completions() -> list[str]:
    """Return a list of available command paths, including builtins."""

    commands: list[str] = []

    # builtins appear as top-level commands
    for name in gw.builtins():
        commands.append(name.replace("_", "-"))

    # discover project functions recursively
    for proj in gw.projects():
        try:
            project = gw.load_project(proj)
        except Exception:
            continue
        for parts in _walk(project, [proj]):
            commands.append(" ".join(p.replace("_", "-") for p in parts))

    return sorted(commands)




# file: tests/test_sigils.py

import unittest
from gway import Sigil, Spool, gw, __


class SigilTests(unittest.TestCase):

    def test_basic_resolution_with_dict(self):
        data = {"name": "Alice"}
        s = Sigil("Hello [name]")
        self.assertEqual(s % data, "Hello Alice")

    def test_unresolved_key_raises(self):
        data = {}
        s = Sigil("Hello [user]")
        with self.assertRaises(KeyError):
            _ = s % data

    def test_case_insensitive_lookup(self):
        data = {"USER": "admin"}
        s = Sigil("Logged in as [user]")
        # Your code tries lower/upper/case variants
        self.assertEqual(s % data, "Logged in as admin")

    def test_multiple_sigils_in_text(self):
        data = {"x": 10, "y": 20}
        s = Sigil("Coordinates: [x], [y]")
        self.assertEqual(s % data, "Coordinates: 10, 20")

    def test_quoted_literal_returns_literal(self):
        s = Sigil('Key is ["LITERAL"]')
        self.assertEqual(s % {}, 'Key is LITERAL')

    def test_unresolved_unquoted_raises(self):
        s = Sigil("Oops [nope]")
        with self.assertRaises(KeyError):
            _ = s % {}

    def test_list_sigils(self):
        s = Sigil("A [foo] and [bar]")
        self.assertEqual(s.list_sigils(), ["[foo]", "[bar]"])

    def test_repr_and_str(self):
        s = Sigil("[val]")
        self.assertIn("[val]", str(s))
        self.assertIn("[val]", repr(s))

    def test_make_lookup_with_callable_variants(self):
        def finder(key, _):
            if key == "FOO_BAR":
                return "bar"
        s = Sigil("Value [FOO-BAR]")
        self.assertEqual(s.resolve(finder), "Value bar")

    def test_dotted_and_spaced_paths(self):
        class Obj:
            pass
        obj = Obj()
        obj.name = "Widget"
        data = {"app": {"name": "dict"}, "obj": obj}
        self.assertEqual(Sigil("[app.name]") % data, "dict")
        self.assertEqual(Sigil("[app name]") % data, "dict")
        self.assertEqual(Sigil("[obj.name]") % data, "Widget")

    def test_standalone_and_embedded_non_string(self):
        data = {"num": 7, "info": {"a": 1}}
        # Standalone should return raw value
        self.assertEqual(Sigil("[num]") % data, 7)
        self.assertEqual(Sigil("[info]") % data, {"a": 1})
        # Embedded should stringify/serialize
        from json import dumps
        self.assertEqual(Sigil("Value [num]") % data, "Value 7")
        self.assertEqual(Sigil("Info=[info]") % data, "Info=" + dumps({"a": 1}))

    def test_underscore_path_denied(self):
        data = {"obj": {"_secret": 42}, "_root": {"val": 1}}
        with self.assertRaises(KeyError):
            Sigil("[obj._secret]") % data
        with self.assertRaises(KeyError):
            Sigil("[obj _secret]") % data
        # Accessing leading underscore as first segment is allowed
        self.assertEqual(Sigil("[_root.val]") % data, 1)

    def test_unquote_helper(self):
        from gway.sigils import _unquote
        self.assertEqual(_unquote('"hello"'), "hello")
        self.assertEqual(_unquote("'world'"), "world")
        self.assertEqual(_unquote("plain"), "plain")

class SpoolTests(unittest.TestCase):
    def setUp(self):
        self.mapping = {"A": "apple", "B": "banana", "C": "cucumber"}

    def test_spool_basic_resolve(self):
        spool = Spool("[B]", "[A]", "[C]")
        # Should resolve the *first* that is found (so B="banana")
        self.assertEqual(spool.resolve(self.mapping), "banana")

    def test_spool_with_literal_and_missing(self):
        spool = Spool("foo", '[B]', '["literal"]')
        # "foo" becomes Sigil("foo"), so will raise. Next, [B] will resolve.
        self.assertEqual(spool.resolve(self.mapping), "foo")

    def test_spool_resolve_raises_when_none_resolved(self):
        spool = Spool("[X]", "[Y]")
        with self.assertRaises(KeyError):
            spool.resolve(self.mapping)

    def test_spool_append_and_extend(self):
        spool = Spool("[A]")
        spool.append("[B]")
        spool.extend(["[C]"])
        self.assertEqual([str(x) for x in spool], ["[A]", "[B]", "[C]"])

    def test_spool_sequence_protocol(self):
        spool = Spool("[A]", "[B]")
        self.assertEqual(len(spool), 2)
        self.assertEqual(str(spool[0]), "[A]")
        self.assertEqual([str(s) for s in spool], ["[A]", "[B]"])

    def test_spool_resolve_with_resolver_object(self):
        spool = Spool("[X]", "[A]")
        # Provide a gw-style object with .resolve()
        self.assertEqual(spool.resolve(self.mapping), "apple")

    def test_spool_flatten_and_str(self):
        spool = Spool(["[A]", ["[B]", "[C]"]])
        self.assertEqual(len(spool), 3)
        self.assertEqual(str(spool), "[A] | [B] | [C]")

if __name__ == "__main__":
    unittest.main()



# file: projects/games/conway.py

import os
import random
import html
from datetime import datetime

from gway import gw
from bottle import response, redirect


BOARD_SIZE = 54
BOARD_FILE = gw.resource("work", "shared", "games","conway.txt", touch=True)

def _new_board(size=BOARD_SIZE, fill=0):
    return [[fill for _ in range(size)] for _ in range(size)]

def _random_board(size=BOARD_SIZE):
    return [[random.choice([0, 1]) for _ in range(size)] for _ in range(size)]

def _serialize_board(board):
    return "\n".join(",".join(str(cell) for cell in row) for row in board)

def _deserialize_board(s):
    return [[int(cell) for cell in row.split(",")] for row in s.strip().splitlines()]

def load_board():
    """Load the board from disk, or create one if missing or empty."""
    # Check for existence and non-empty file
    if not os.path.exists(BOARD_FILE) or os.path.getsize(BOARD_FILE) == 0:
        board = _random_board()
        save_board(board)
        return board
    with open(BOARD_FILE, "r", encoding="utf-8") as f:
        try:
            return _deserialize_board(f.read())
        except Exception:
            # On any error (e.g., file corrupt), start fresh
            board = _random_board()
            save_board(board)
            return board

def save_board(board):
    """Save the board to disk as CSV rows."""
    with open(BOARD_FILE, "w", encoding="utf-8") as f:
        f.write(_serialize_board(board))

def is_board_empty(board):
    return all(cell == 0 for row in board for cell in row)

def is_board_full(board):
    return all(cell == 1 for row in board for cell in row)

def flip_board(board):
    return [[1 - cell for cell in row] for row in board]

def next_generation(board):
    size = len(board)
    def neighbors(r, c):
        return sum(
            board[(r+dr)%size][(c+dc)%size]
            for dr in (-1,0,1) for dc in (-1,0,1)
            if (dr,dc)!=(0,0)
        )
    return [
        [1 if (cell and 2<=neighbors(r,c)<=3) or (not cell and neighbors(r,c)==3) else 0
         for c,cell in enumerate(row)]
        for r,row in enumerate(board)
    ]

def view_game_of_life(
    *args,
    action=None,
    board=None,
    toggle_x=None,
    toggle_y=None,
    **kwargs
):
    # 1. Load the current board
    msg = ""
    current_board = load_board()

    # 2. Handle form board (from hidden input)
    if board:
        try:
            new_board = [
                [int(cell) for cell in row.split(",")]
                for row in board.strip().split(";")
            ]
            current_board = new_board
        except Exception:
            msg = "Failed to parse board!"

    # 3. Action handlers
    if action == "random":
        current_board = _random_board()
    elif action == "clear":
        current_board = _new_board()
    elif action == "step":
        current_board = next_generation(current_board)
    elif action == "toggle":
        x = int(toggle_x) if toggle_x is not None else -1
        y = int(toggle_y) if toggle_y is not None else -1
        if 0 <= x < BOARD_SIZE and 0 <= y < BOARD_SIZE:
            current_board[x][y] = 0 if current_board[x][y] else 1
        save_board(current_board)
        return redirect("/conway/game-of-life")  # POST-redirect-GET

    # 4. Always save board for step/random/clear (toggle already saved above)
    if action in ("step", "random", "clear"):
        save_board(current_board)

    # 5. Prepare HTML board
    html_board = ""
    for x, row in enumerate(current_board):
        row_html = "".join(
            f'<td class="cell cell-{cell}" data-x="{x}" data-y="{y}"></td>'
            for y, cell in enumerate(row)
        )
        html_board += f"<tr>{row_html}</tr>"

    flat_board = ";".join(",".join(str(cell) for cell in row) for row in current_board)

    # 6. Get last modified date of board file
    last_mod = None
    if os.path.exists(BOARD_FILE):
        last_mod = datetime.fromtimestamp(os.path.getmtime(BOARD_FILE)).strftime("%Y-%m-%d %H:%M:%S")

    ICONS = {
        "step": '<svg viewBox="0 0 20 20"><polyline points="5,3 15,10 5,17" fill="none" stroke="currentColor" stroke-width="2"/></svg>',
        "random": '<svg viewBox="0 0 20 20"><circle cx="10" cy="10" r="8" fill="none" stroke="currentColor" stroke-width="2"/><circle cx="10" cy="10" r="3" fill="currentColor"/></svg>',
        "clear": '<svg viewBox="0 0 20 20"><rect x="4" y="4" width="12" height="12" fill="none" stroke="currentColor" stroke-width="2"/><line x1="6" y1="6" x2="14" y2="14" stroke="currentColor" stroke-width="2"/><line x1="14" y1="6" x2="6" y2="14" stroke="currentColor" stroke-width="2"/></svg>',
        "download": '<svg viewBox="0 0 20 20"><path d="M10 3v10m0 0l-4-4m4 4l4-4M3 17h14" fill="none" stroke="currentColor" stroke-width="2"/></svg>',
    }

    return f"""
    <h1>Conway's Game of Life</h1>
    <div>
        <form id="lifeform" method="post" class="game-actions" autocomplete="off" style="margin-bottom:8px;">
            <input type="hidden" name="board" id="boarddata" value="{html.escape(flat_board)}" />
            <button type="submit" name="action" value="step">{ICONS['step']} Step Forward</button>
            <button type="submit" name="action" value="random">{ICONS['random']} Random</button>
            <button type="submit" name="action" value="clear">{ICONS['clear']} Clear</button>
            <a href="/shared/games/conway.txt" download class="button">{ICONS['download']} Download</a>
        </form>
        <p style="color:#aa2222">{html.escape(msg)}</p>
        <table id="gameboard" class="game-board">{html_board}</table>
        <div style="margin-top:0.8em;color:#888;font-size:0.96em">
            Last updated: {last_mod or "Never"}
        </div>
    </div>
    """



# file: projects/monitor/monitor.py

import time
import datetime
import asyncio
from gway import gw

NETWORK_STATE = {}         # {project: {last_run, last_result, ...}}
MONITOR_NEXT_CHECK = {}    # {project: next_check_iso}
MONITOR_TRIGGER = {}       # {project: asyncio.Event (optional)}
MONITOR_RENDER = {}        # {project: list of renderers}

def now_iso():
    """Return current time as ISO8601 string (seconds precision)."""
    return datetime.datetime.now().isoformat(timespec="seconds")

def get_state(project):
    """Return the state dict for the given project, creating it if needed."""
    project = str(project)
    return NETWORK_STATE.setdefault(project, {})

def set_state(project, key, value):
    """Set a single key/value in the state for a given project."""
    get_state(project)[key] = value

def set_states(project, mapping):
    """Set multiple key/values in the state for a given project."""
    get_state(project).update(mapping)

def get_next_check_time(project):
    """Get the next scheduled check time for a given project (ISO8601 or None)."""
    return MONITOR_NEXT_CHECK.get(project)

def trigger_watch(project):
    """Trigger an immediate check for a given project, if async Event is set up."""
    trig = MONITOR_TRIGGER.get(project)
    if trig and hasattr(trig, "set"):
        trig.set()
        return True
    return False

def start_watch(
    project, *,
    monitor=None,
    interval=120,
    delay=0,
    block=False,
    daemon=True,
    render=None,
    logger=None,
    **kwargs
):
    """
    Start a watcher loop for a project.
      - project: Name of the GWAY project or subproject.
      - monitor: Name or list of monitor functions (string or list, without prefix).
      - interval: Minimum seconds between checks.
      - delay: Startup delay (seconds).
      - block: Block main thread? (default False)
      - daemon: Async coroutine? (default True)
      - render: Name or list of render functions (without prefix), for dashboard.
      - logger: Optional logger (uses print if not provided).
      - kwargs: Extra parameters for monitor functions.
    """
    global MONITOR_NEXT_CHECK, MONITOR_TRIGGER, MONITOR_RENDER

    project_name = str(project)
    log_prefix = f"[monitor:{project_name}] "

    gproj = gw.find_project(f"monitor.{project_name}")
    if not gproj:
        raise ValueError(f"{log_prefix}Project not found in GWAY: 'monitor.{project_name}'")

    # Select monitors (function names)
    monitors = gw.cast.to_list(monitor) if monitor else [project_name]
    monitor_funcs = []
    for mname in monitors:
        funcname = f"monitor_{mname}"
        func = getattr(gproj, funcname, None)
        if not func:
            raise ValueError(f"{log_prefix}Monitor function '{funcname}' not found in project '{project_name}'.")
        monitor_funcs.append((funcname, func))

    # Set up renders for this project (for dashboard view)
    if render is not None:
        renders = gw.cast.to_list(render)
        MONITOR_RENDER[project_name] = renders
    elif project_name not in MONITOR_RENDER:
        MONITOR_RENDER[project_name] = [project_name]

    def log_info(msg):
        if logger: logger.info(msg)
        else: print(f"{log_prefix}{msg}")

    def log_warn(msg):
        if logger: logger.warning(msg)
        else: print(f"{log_prefix}[WARN] {msg}")

    trig = None

    async def async_loop():
        """Async watcher loop (recommended for web/production)."""
        nonlocal trig
        if delay > 0:
            log_info(f"Waiting {delay}s before starting monitor loop...")
            await asyncio.sleep(delay)
        trig = MONITOR_TRIGGER.setdefault(project_name, asyncio.Event())
        while True:
            state = get_state(project_name)
            state["last_run"] = now_iso()
            state["last_result"] = results = []
            for funcname, func in monitor_funcs:
                try:
                    log_info(f"Calling {funcname} ...")
                    result = func(**kwargs)
                    results.append((funcname, result))
                except Exception as e:
                    log_warn(f"Exception in {funcname}: {e}")
                    results.append((funcname, f"error: {e}"))
            next_time = datetime.datetime.now() + datetime.timedelta(seconds=interval)
            MONITOR_NEXT_CHECK[project_name] = next_time.isoformat(timespec="seconds")
            try:
                await asyncio.wait_for(trig.wait(), timeout=interval)
                trig.clear()
                log_info("Triggered immediate check via trigger_watch.")
            except asyncio.TimeoutError:
                pass

    def blocking_loop():
        """Synchronous watcher loop (for legacy CLI or quick tests)."""
        if delay > 0:
            log_info(f"Waiting {delay}s before starting monitor loop...")
            time.sleep(delay)
        while True:
            state = get_state(project_name)
            state["last_run"] = now_iso()
            state["last_result"] = results = []
            for funcname, func in monitor_funcs:
                try:
                    log_info(f"Calling {funcname} ...")
                    result = func(**kwargs)
                    results.append((funcname, result))
                except Exception as e:
                    log_warn(f"Exception in {funcname}: {e}")
                    results.append((funcname, f"error: {e}"))
            next_time = datetime.datetime.now() + datetime.timedelta(seconds=interval)
            MONITOR_NEXT_CHECK[project_name] = next_time.isoformat(timespec="seconds")
            time.sleep(interval)

    # Main entry points
    if daemon:
        return async_loop()
    if block:
        blocking_loop()
    else:
        state = get_state(project_name)
        state["last_run"] = now_iso()
        state["last_result"] = results = []
        for funcname, func in monitor_funcs:
            try:
                log_info(f"Calling {funcname} ...")
                result = func(**kwargs)
                results.append((funcname, result))
            except Exception as e:
                log_warn(f"Exception in {funcname}: {e}")
                results.append((funcname, f"error: {e}"))
        next_time = datetime.datetime.now() + datetime.timedelta(seconds=interval)
        MONITOR_NEXT_CHECK[project_name] = next_time.isoformat(timespec="seconds")
        return results

def view_monitor_panel(**_):
    """
    Dashboard view: Displays current state and HTML reports for each configured monitor.
    Each monitor section is rendered by calling the specified render function(s) in the project:
      - render_<render> for each render in MONITOR_RENDER[project]
      - fallback: render_<project> or render_monitor
    Below the monitor, display last run and next scheduled run.
    """
    html = ['<div class="gway-net-dashboard">']
    if not NETWORK_STATE:
        html.append('<div class="warn" style="color:#a00;">No monitors are currently running.</div>')
    for project in NETWORK_STATE:
        state = get_state(project)
        html.append(f'<div class="monitor-block">')
        gproj = gw.find_project(f"monitor.{project}")
        renders = MONITOR_RENDER.get(project) or [project]
        rendered = False
        for rname in renders:
            funcname = f"render_{rname}"
            func = getattr(gproj, funcname, None)
            if func:
                try:
                    frag = func()
                    html.append(f'<div class="monitor-frag">{frag}</div>')
                    rendered = True
                except Exception as e:
                    html.append(f'<div class="error">Error in {funcname}: {e}</div>')
        if not rendered:
            for fallback in [f"render_{project}", "render_monitor"]:
                func = getattr(gproj, fallback, None)
                if func:
                    try:
                        frag = func()
                        html.append(f'<div class="monitor-frag">{frag}</div>')
                        rendered = True
                        break
                    except Exception as e:
                        html.append(f'<div class="error">Error in {fallback}: {e}</div>')
        if not rendered:
            html.append('<div class="no-render" style="color:#888;">No render function found for this monitor.</div>')
        # Show info about last run and next run
        last_run = state.get("last_run") or "-"
        next_run = get_next_check_time(project) or "-"
        html.append(f'<div class="monitor-meta" style="margin-top:7px;font-size:90%;color:#888;">'
                    f'Last run: <b>{last_run}</b> | Next check: <b>{next_run}</b></div>')
        html.append('</div>')
    html.append('</div>')
    return '\n'.join(html)



import unittest
import base64
from gway import gw

web = gw.load_project('web')


class ParseBasicAuthHeaderTests(unittest.TestCase):
    def test_valid_header(self):
        username = 'foo'
        password = 'bar'
        creds = f"{username}:{password}".encode('utf-8')
        header = 'Basic ' + base64.b64encode(creds).decode()
        self.assertEqual(
            web.auth.parse_basic_auth_header(header),
            (username, password),
        )

    def test_invalid_headers(self):
        invalid_headers = [
            None,
            '',
            'Bearer abc',
            'Basic not-base64',
            'Basic ' + base64.b64encode(b'nocolon').decode(),
        ]
        for hdr in invalid_headers:
            with self.subTest(hdr=hdr):
                self.assertEqual(
                    web.auth.parse_basic_auth_header(hdr),
                    (None, None),
                )


if __name__ == '__main__':
    unittest.main()



# file: projects/ocpp/csms.py
# web.app path: ocpp/csms/

import json
import os
import shutil
import time
import uuid
import traceback
import asyncio
import html
from datetime import datetime, timedelta
from fastapi import WebSocket, WebSocketDisconnect
from bottle import request, redirect, HTTPError
from typing import Dict, Optional
from gway import gw

_csms_loop: Optional[asyncio.AbstractEventLoop] = None
_transactions: Dict[str, dict] = {}
_active_cons: Dict[str, WebSocket] = {}
_latest_heartbeat: Dict[str, str] = {}
_abnormal_status: Dict[str, dict] = {}
_msg_log: Dict[str, list] = {}

def bind_state(root):
    """Bind shared dictionaries from ``ocpp`` root module."""
    global _transactions, _active_cons, _latest_heartbeat, _abnormal_status, _msg_log
    _transactions = root._transactions
    _active_cons = root._active_cons
    _latest_heartbeat = root._latest_heartbeat
    _abnormal_status = root._abnormal_status
    _msg_log = root._msg_log

def authorize_balance(**record):
    """
    Default OCPP RFID secondary validator: Only authorize if balance >= 1.
    The RFID needs to exist already for this to be called in the first place.
    """
    try:
        return float(record.get("balance", "0")) >= 1
    except Exception:
        return False
    
def setup_app(*,
    app=None,
    allowlist=None,
    denylist=None,
    location=None,
    authorize=authorize_balance,
    email=None,
    auth="disabled",
):
    # no globals needed here; dictionaries are modified in-place
    email = email if isinstance(email, str) else (gw.resolve('[ADMIN_EMAIL]') if email else email)

    auth_required = str(auth).strip().lower() not in {
        "none", "false", "disabled", "optional"
    }

    oapp = app
    from fastapi import FastAPI as _FastAPI
    match app:
        case _FastAPI() as f:
            app = f
            _is_new_app = False
        case list() | tuple() as seq:
            app = next((x for x in seq if isinstance(x, _FastAPI)), None)
            _is_new_app = app is None
        case None:
            _is_new_app = True
        case _ if isinstance(app, _FastAPI):
            _is_new_app = False
        case _ if hasattr(app, "__iter__") and not isinstance(app, (str, bytes, bytearray)):
            app = next((x for x in app if isinstance(x, _FastAPI)), None)
            _is_new_app = app is None
        case _:
            _is_new_app = app is None or not isinstance(app, _FastAPI)
    if _is_new_app:
        app = _FastAPI()

    validator = None
    if isinstance(authorize, str):
        validator = gw[authorize]
    elif callable(authorize):
        validator = authorize

    def is_authorized_rfid(rfid: str) -> bool:
        if denylist and gw.cdv.validate(denylist, rfid):
            gw.info(f"[OCPP] RFID {rfid!r} is present in denylist. Authorization denied.")
            return False
        if not allowlist:
            gw.warn("[OCPP] No RFID allowlist configured — rejecting all authorization requests.")
            return False
        return gw.cdv.validate(allowlist, rfid, validator=validator)

    @app.websocket("/{path:path}")
    async def websocket_ocpp(websocket: WebSocket, path: str):
        global _csms_loop
        if auth_required:
            if not gw.web.auth.check_websocket_auth(websocket):
                await websocket.close(code=4401, reason="Unauthorized")
                gw.warn(f"[OCPP] Unauthorized WebSocket connection attempt for charger_id={path}")
                return

        _csms_loop = asyncio.get_running_loop()
        charger_id = path.strip("/").split("/")[-1]
        gw.info(f"[OCPP] WebSocket connected: charger_id={charger_id}")

        protos = websocket.headers.get("sec-websocket-protocol", "").split(",")
        protos = [p.strip() for p in protos if p.strip()]
        if "ocpp1.6" in protos:
            await websocket.accept(subprotocol="ocpp1.6")
        else:
            await websocket.accept()

        _active_cons[charger_id] = websocket

        try:
            while True:
                raw = await websocket.receive_text()
                gw.info(f"[OCPP:{charger_id}] → {raw}")
                _msg_log.setdefault(charger_id, []).append(f"> {raw}")
                try:
                    msg = json.loads(raw)
                except json.JSONDecodeError:
                    gw.warn(f"[OCPP:{charger_id}] Received non-JSON message: {raw!r}")
                    continue

                if isinstance(msg, list) and msg[0] == 2:
                    message_id, action = msg[1], msg[2]
                    payload = msg[3] if len(msg) > 3 else {}
                    gw.debug(f"[OCPP:{charger_id}] Action={action} Payload={payload}")

                    response_payload = {}

                    if action == "Authorize":
                        status = "Accepted" if is_authorized_rfid(payload.get("idTag")) else "Rejected"
                        response_payload = {"idTagInfo": {"status": status}}

                    elif action == "BootNotification":
                        response_payload = {
                            "currentTime": datetime.utcnow().isoformat() + "Z",
                            "interval": 300,
                            "status": "Accepted"
                        }

                    elif action == "Heartbeat":
                        response_payload = {"currentTime": datetime.utcnow().isoformat() + "Z"}

                    elif action == "StartTransaction":
                        now = int(time.time())
                        transaction_id = now
                        _transactions[charger_id] = {
                            "syncStart": 1,
                            "connectorId": payload.get("connectorId"),
                            "idTagStart": payload.get("idTag"),
                            "meterStart": payload.get("meterStart"),
                            "reservationId": payload.get("reservationId", -1),
                            "startTime": now,
                            "startTimeStr": datetime.utcfromtimestamp(now).isoformat() + "Z",
                            "startMs": int(time.time() * 1000) % 1000,
                            "transactionId": transaction_id,
                            "MeterValues": []
                        }
                        cp_ts = None
                        if payload.get("timestamp"):
                            try:
                                cp_ts = int(datetime.fromisoformat(payload["timestamp"].rstrip("Z")).timestamp())
                            except Exception:
                                cp_ts = None
                        gw.ocpp.data.record_transaction_start(
                            charger_id,
                            transaction_id,
                            now,
                            id_tag=payload.get("idTag"),
                            meter_start=payload.get("meterStart"),
                            charger_timestamp=cp_ts,
                        )
                        response_payload = {
                            "transactionId": transaction_id,
                            "idTagInfo": {"status": "Accepted"}
                        }

                        if email:
                            subject = f"OCPP: Charger {charger_id} STARTED transaction {transaction_id}"
                            body = (
                                f"Charging session started.\n"
                                f"Charger: {charger_id}\n"
                                f"idTag: {payload.get('idTag')}\n"
                                f"Connector: {payload.get('connectorId')}\n"
                                f"Start Time: {datetime.utcfromtimestamp(now).isoformat()}Z\n"
                                f"Transaction ID: {transaction_id}\n"
                                f"Meter Start: {payload.get('meterStart')}\n"
                                f"Reservation ID: {payload.get('reservationId', -1)}"
                            )
                            gw.mail.send(subject, body, to=email)

                    elif action == "MeterValues":
                        tx = _transactions.get(charger_id)
                        if tx:
                            for entry in payload.get("meterValue", []):
                                ts = entry.get("timestamp")
                                ts_epoch = (
                                    int(datetime.fromisoformat(ts.rstrip("Z")).timestamp())
                                    if ts else int(time.time())
                                )
                                sampled = []
                                for sv in entry.get("sampledValue", []):
                                    val = sv.get("value")
                                    unit = sv.get("unit", "")
                                    measurand = sv.get("measurand", "")
                                    try:
                                        fval = float(val)
                                        if unit == "Wh":
                                            fval = fval / 1000.0
                                        sampled.append({
                                            "value": fval,
                                            "unit": "kWh" if unit == "Wh" else unit,
                                            "measurand": measurand,
                                            "context": sv.get("context", ""),
                                        })
                                        gw.ocpp.data.record_meter_value(
                                            charger_id,
                                            tx.get("transactionId"),
                                            ts_epoch,
                                            measurand,
                                            fval,
                                            "kWh" if unit == "Wh" else unit,
                                            sv.get("context", ""),
                                        )
                                    except Exception:
                                        continue
                                tx["MeterValues"].append({
                                    "timestamp": ts_epoch,
                                    "timestampStr": datetime.utcfromtimestamp(ts_epoch).isoformat() + "Z",
                                    "timeMs": int(time.time() * 1000) % 1000,
                                    "sampledValue": sampled,
                                })
                        response_payload = {}

                    elif action == "StopTransaction":
                        now = int(time.time())
                        tx = _transactions.get(charger_id)
                        if tx:
                            if tx.get("MeterValues"):
                                try:
                                    archive_energy(charger_id, tx["transactionId"], tx["MeterValues"])
                                except Exception as e:
                                    gw.error("Error recording energy chart.")
                                    gw.exception(e)
                            tx.update({
                                "syncStop": 1,
                                "idTagStop": payload.get("idTag"),
                                "meterStop": payload.get("meterStop"),
                                "stopTime": now,
                                "stopTimeStr": datetime.utcfromtimestamp(now).isoformat() + "Z",
                                "stopMs": int(time.time() * 1000) % 1000,
                                "reason": 4,
                                "reasonStr": "Local",
                            })
                            cp_stop = None
                            if payload.get("timestamp"):
                                try:
                                    cp_stop = int(datetime.fromisoformat(payload["timestamp"].rstrip("Z")).timestamp())
                                except Exception:
                                    cp_stop = None
                            gw.ocpp.data.record_transaction_stop(
                                charger_id,
                                tx.get("transactionId"),
                                now,
                                meter_stop=payload.get("meterStop"),
                                reason="Local",
                                charger_timestamp=cp_stop,
                            )
                            archive_transaction(charger_id, tx)
                            if location:
                                file_path = gw.resource(
                                    "work", "etron", "records", location,
                                    f"{charger_id}_{tx['transactionId']}.dat"
                                )
                                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                                with open(file_path, "w") as f:
                                    json.dump(tx, f, indent=2)
                        response_payload = {"idTagInfo": {"status": "Accepted"}}

                    elif action == "StatusNotification":
                        status = payload.get("status")
                        error_code = payload.get("errorCode")
                        info = payload.get("info", "")
                        # Only store if abnormal; remove if cleared
                        if is_abnormal_status(status, error_code):
                            _abnormal_status[charger_id] = {
                                "status": status,
                                "errorCode": error_code,
                                "info": info,
                                "timestamp": datetime.utcnow().isoformat() + "Z"
                            }
                            gw.warn(f"[OCPP] Abnormal status for {charger_id}: {status}/{error_code} - {info}")
                            gw.ocpp.data.record_error(charger_id, status, error_code, info)
                        else:
                            if charger_id in _abnormal_status:
                                gw.info(f"[OCPP] Status normalized for {charger_id}: {status}/{error_code}")
                                _abnormal_status.pop(charger_id, None)
                        response_payload = {}

                    else:
                        response_payload = {"status": "Accepted"}

                    response = [3, message_id, response_payload]
                    gw.info(f"[OCPP:{charger_id}] ← {action} => {response_payload}")
                    await websocket.send_text(json.dumps(response))

                elif isinstance(msg, list) and msg[0] == 3:
                    # Handle CALLRESULT, check for Heartbeat ACK to record latest heartbeat time
                    payload = msg[2] if len(msg) > 2 else {}
                    if isinstance(payload, dict) and "currentTime" in payload:
                        # Only update for Heartbeat (or any other call with currentTime)
                        _latest_heartbeat[charger_id] = payload["currentTime"]
                        gw.debug(f"[OCPP:{charger_id}] Updated latest heartbeat to {_latest_heartbeat[charger_id]}")
                    continue

                elif isinstance(msg, list) and msg[0] == 4:
                    gw.info(f"[OCPP:{charger_id}] Received CALLERROR: {msg}")
                    continue

                else:
                    gw.warn(f"[OCPP:{charger_id}] Invalid or unsupported message format: {msg}")

        except WebSocketDisconnect:
            gw.info(f"[OCPP:{charger_id}] WebSocket disconnected")
        except Exception as e:
            gw.error(f"[OCPP:{charger_id}] WebSocket failure: {e}")
            gw.debug(traceback.format_exc())
        finally:
            _active_cons.pop(charger_id, None)

    return (app if not oapp else (oapp, app)) if _is_new_app else oapp

...

def is_abnormal_status(status: str, error_code: str) -> bool:
    """Determine if a status/errorCode is 'abnormal' per OCPP 1.6."""
    status = (status or "").capitalize()
    error_code = (error_code or "").capitalize()
    # Available/NoError or Preparing are 'normal'
    if status in ("Available", "Preparing") and error_code in ("Noerror", "", None):
        return False
    # All Faulted, Unavailable, Suspended, etc. are abnormal
    if status in ("Faulted", "Unavailable", "Suspendedev", "Suspended", "Removed"):
        return True
    if error_code not in ("Noerror", "", None):
        return True
    return False

def get_charger_state(cid, tx, ws_live, raw_hb):
    """
    Determine charger state for stripe:
    - "error": charger is abnormal/faulted
    - "online": live socket, active transaction, not closed
    - "available": live socket, no open transaction
    - "unknown": not live, not abnormal, default
    """
    # Priority: error > online > available > unknown
    if cid in _abnormal_status:
        return "error"
    if ws_live and tx and not tx.get("syncStop"):
        return "online"
    if ws_live and (not tx or tx.get("syncStop")):
        return "available"
    return "unknown"

...

def _render_card_controls(cid):
    return f'''
            <form method="post" action="" class="charger-action-form">
              <input type="hidden" name="charger_id" value="{cid}">
              <select name="action" id="action-{cid}" aria-label="Action">
                <option value="remote_stop">Stop</option>
                <option value="reset_soft">Soft Reset</option>
                <option value="reset_hard">Hard Reset</option>
                <option value="disconnect">Disconnect</option>
              </select>
              <div class="charger-actions-btns">
                <button type="submit" name="do" value="send">Send</button>
                <a href="/ocpp/csms/energy-graph?charger_id={cid}" class="graph-btn" target="_blank">Graph</a>
              </div>
            </form>
    '''

def _render_card_link(cid):
    return (
        '<div class="charger-actions-btns">'
        f'<a href="/ocpp/csms/charger-detail?charger_id={cid}" class="graph-btn">Details</a>'
        '</div>'
    )

def _render_charger_card(cid, tx, state, raw_hb, *, show_controls=True):
    """
    Render a charger card with the right status stripe (state: "online", "available", "error", "unknown").
    """
    status_class = f"status-{state}"
    tx_id       = tx.get("transactionId") if tx else '-'
    meter_start = tx.get("meterStart") if tx else '-'
    latest      = (
        tx.get("meterStop")
        if tx and tx.get("meterStop") is not None
        else (tx["MeterValues"][-1].get("meter") if tx and tx.get("MeterValues") else 'None')
    )
    power  = power_consumed(tx)
    status = "Closed" if tx and tx.get("syncStop") else "Open" if tx else '-'
    latest_hb = "-"
    if raw_hb:
        try:
            from datetime import datetime
            dt = datetime.fromisoformat(raw_hb.replace("Z", "+00:00")).astimezone()
            latest_hb = dt.strftime("%Y-%m-%d %H:%M:%S")
        except Exception:
            latest_hb = raw_hb

    last_updated = tx.get("last_updated", raw_hb or "") if tx else (raw_hb or "")

    controls_html = _render_card_controls(cid) if show_controls else _render_card_link(cid)
    details_html = ''

    return f'''
    <div class="charger-card {status_class}" id="charger-{cid}">
      <input type="hidden" name="last_updated" value="{last_updated}">
      <table class="charger-layout">
        <tr>
          <td class="charger-info-td">
            <table class="charger-info-table">
              <tr>
                <td class="label">ID</td>
                <td class="value">{cid}</td>
              </tr>
              <tr>
                <td class="label">TXN</td>
                <td class="value">{tx_id}</td>
              </tr>
              <tr>
                <td class="label">Start</td>
                <td class="value">{meter_start}</td>
                <td class="label">Latest</td>
                <td class="value">{latest}</td>
              </tr>
              <tr>
                <td class="label">kWh.</td>
                <td class="value">{power}</td>
                <td class="label">Last HB</td>
                <td class="value">{latest_hb}</td>
              </tr>
              <tr>
                <td class="label">Status</td>
                <td class="value" colspan="3">{status}</td>
              </tr>
            </table>
          </td>
          <td class="charger-actions-td">
            {controls_html}
          </td>
        </tr>
      </table>
      {details_html}
    </div>
    '''

def view_charger_status(*, action=None, charger_id=None, show=None, **_):
    """
    Card-based OCPP dashboard: summary of charger connections.
    Renders <div id="charger-list" data-gw-render="charger_list" data-gw-refresh="5">
    so the client can periodically refresh the list via render.js.
    ``show=all`` includes historic chargers from the database.
    """
    msg = ""
    show = show or request.query.get("show")
    if request.method == "POST":
        action = request.forms.get("action")
        charger_id = request.forms.get("charger_id")
        if action and charger_id:
            try:
                dispatch_action(charger_id, action)
                msg = f"Action {action} sent"
            except Exception as e:
                gw.error(f"Failed to dispatch action {action} to {charger_id}: {e}")
                msg = f"Error: {e}"

    all_chargers = set(_active_cons) | set(_transactions)
    if show == "all":
        try:
            all_chargers |= set(gw.ocpp.data.list_chargers())
        except Exception:
            pass
    html = [
        '<link rel="stylesheet" href="/static/ocpp/csms/charger_status.css">',
        '<script src="/static/render.js"></script>',
        '<script src="/static/ocpp/csms/charger_status.js"></script>',
        "<h1>OCPP Status Dashboard</h1>"
    ]
    if msg:
        html.append(f'<p class="error">{msg}</p>')

    link_all = '<a href="?show=all">Show all offline chargers</a>'
    link_new = '<a href="?show=new">Show active and recent chargers</a>'
    html.append(f"<p>{link_new if show == 'all' else link_all}</p>")

    # Abnormal status warning
    if _abnormal_status:
        html.append(
            '<div class="ocpp-alert">'
            "⚠️ Abnormal Charger Status Detected:<ul>"
        )
        for cid, err in sorted(_abnormal_status.items()):
            status = err.get("status", "")
            error_code = err.get("errorCode", "")
            info = err.get("info", "")
            ts = err.get("timestamp", "")
            msg = f"<b>{cid}</b>: {status}/{error_code}"
            if info: msg += f" ({info})"
            if ts: msg += f" <span class='timestamp'>@{ts}</span>"
            html.append(f"<li>{msg}</li>")
        html.append("</ul></div>")

    # --- The key block for autorefresh ---
    html.append(
        f'<div id="charger-list" data-gw-render="charger_list" data-gw-refresh="5" data-gw-click="refresh" data-show="{show or ""}">' 
    )
    if not all_chargers:
        html.append('<p><em>No chargers connected or transactions seen yet.</em></p>')
    else:
        for cid in sorted(all_chargers):
            ws_live = cid in _active_cons
            tx = _transactions.get(cid)
            raw_hb = _latest_heartbeat.get(cid)
            state = get_charger_state(cid, tx, ws_live, raw_hb)
            html.append(_render_charger_card(cid, tx, state, raw_hb, show_controls=False))
    html.append('</div>')

    ws_url = gw.web.build_ws_url()
    html.append(f"""
    <div class="ocpp-wsbar">
      <input type="text" id="ocpp-ws-url" value="{ws_url}" readonly
        style="flex:1;font-family:monospace;font-size:1em;
               padding:10px 6px;background:#222;color:#fff;
               border:1px solid #333;border-radius:5px;min-width:160px;max-width:530px;"/>
      <button id="copy-ws-url-btn"
        style="padding:6px 16px;font-size:1em;border-radius:5px;
               border:1px solid #444;background:#444;color:#fff;cursor:pointer">
        Copy
      </button>
      <a href="/ocpp/evcs/cp-simulator" style="margin-left:1.3em">Simulator</a>
    </div>
    """)
    return "".join(html)

def render_charger_list(*, show=None, **kwargs):
    """
    Regenerate the full charger list HTML (all cards).
    No parsing of incoming HTML; just returns a new block of HTML for charger-list.
    Called via POST (or GET) from render.js, possibly with params in kwargs.
    ``show=all`` includes historic chargers from the database.
    """
    show = show or kwargs.get("show") or request.forms.get("show") or request.query.get("show")
    all_chargers = set(_active_cons) | set(_transactions)
    if show == "all":
        try:
            all_chargers |= set(gw.ocpp.data.list_chargers())
        except Exception:
            pass
    html = []
    if not all_chargers:
        html.append('<p><em>No chargers connected or transactions seen yet.</em></p>')
    else:
        for cid in sorted(all_chargers):
            ws_live = cid in _active_cons
            tx = _transactions.get(cid)
            raw_hb = _latest_heartbeat.get(cid)
            state = get_charger_state(cid, tx, ws_live, raw_hb)
            html.append(_render_charger_card(cid, tx, state, raw_hb, show_controls=False))
    return "\n".join(html)

def view_charger_detail(*, charger_id=None, **_):
    """Detail view for a single charger with live log."""
    if not charger_id:
        return redirect("/ocpp/csms/charger-status")
    known_ids = set(_active_cons) | set(_transactions) | set(_latest_heartbeat)
    if charger_id not in known_ids:
        try:
            if charger_id not in gw.ocpp.data.list_chargers():
                return redirect("/ocpp/csms/charger-status")
        except Exception:
            return redirect("/ocpp/csms/charger-status")

    msg = ""
    if request.method == "POST":
        action = request.forms.get("action")
        if action:
            try:
                dispatch_action(charger_id, action)
                msg = f"Action {action} sent"
            except Exception as e:
                gw.error(f"Failed to dispatch action {action} to {charger_id}: {e}")
                msg = f"Error: {e}"

    ws_live = charger_id in _active_cons
    tx = _transactions.get(charger_id)
    raw_hb = _latest_heartbeat.get(charger_id)
    state = get_charger_state(charger_id, tx, ws_live, raw_hb)
    now = datetime.utcnow()
    since_default = (now - timedelta(days=1)).date().isoformat()
    until_default = now.date().isoformat()
    since = request.query.get('since') or since_default
    until = request.query.get('until') or until_default

    html = [
        '<link rel="stylesheet" href="/static/ocpp/csms/charger_status.css">',
        '<script src="/static/render.js"></script>',
        f'<h1><a href="/ocpp/csms/charger-status">All Chargers</a> / {charger_id} Details</h1>'
    ]
    if msg:
        html.append(f'<p class="error">{msg}</p>')

    html.append(
        f'<div id="charger-info" data-gw-render="charger_info" data-gw-refresh="5" data-gw-click="refresh" data-charger-id="{charger_id}">' +
        _render_charger_card(charger_id, tx, state, raw_hb) +
        '</div>'
    )

    html.append(
        f'''<form id="tx-range" method="get" style="margin:1em 0;">
            <input type="hidden" name="charger_id" value="{charger_id}">
            <label>From: <input type="date" name="since" value="{since}"></label>
            <label>To: <input type="date" name="until" value="{until}"></label>
            <button type="submit">Apply</button>
        </form>'''
    )

    html.append(
        f'<div id="charger-transactions" data-gw-render="charger_transactions" '
        f'data-charger-id="{charger_id}" data-since="{since}" data-until="{until}">' +
        render_charger_transactions(charger_id=charger_id, since=since, until=until) +
        '</div>'
    )

    html.append(
        f'<div id="charger-log" data-gw-render="charger_log" data-gw-refresh="2" data-gw-click="refresh" data-charger-id="{charger_id}">' +
        render_charger_log(charger_id=charger_id) +
        '</div>'
    )

    return "".join(html)

def render_charger_info(*, charger_id=None, chargerId=None, **_):
    cid = charger_id or chargerId
    if not cid:
        return ""
    tx = _transactions.get(cid)
    raw_hb = _latest_heartbeat.get(cid)
    state = get_charger_state(cid, tx, cid in _active_cons, raw_hb)
    return _render_charger_card(cid, tx, state, raw_hb)

def render_charger_log(*, charger_id=None, chargerId=None, **_):
    cid = charger_id or chargerId
    if not cid:
        return ""
    lines = _msg_log.get(cid, [])[-50:]
    esc = lambda s: html.escape(s)
    return '<pre>' + '\n'.join(esc(line) for line in lines) + '</pre>'

def render_charger_transactions(*, charger_id=None, chargerId=None, since=None, until=None, **_):
    cid = charger_id or chargerId
    if not cid:
        return ""
    since = since or request.forms.get('since') or request.query.get('since')
    until = until or request.forms.get('until') or request.query.get('until')

    def to_epoch(date_str):
        if not date_str:
            return None
        try:
            dt = datetime.fromisoformat(date_str)
            if len(date_str) == 10:
                return int(dt.timestamp())
            return int(dt.timestamp())
        except Exception:
            return None

    start_ts = to_epoch(since)
    end_ts = to_epoch(until)
    rows = list(
        gw.ocpp.data.iter_transactions(
            cid,
            start=start_ts,
            end=end_ts,
            sort="start_time",
            order="desc",
            limit=50,
        )
    )

    def fmt(ts):
        if not ts:
            return "-"
        try:
            return datetime.utcfromtimestamp(int(ts)).isoformat() + "Z"
        except Exception:
            return str(ts)

    html = [
        '<table class="ocpp-details">',
        '<tr><th>ID</th><th>Start</th><th>Stop</th><th>Meter Δ(kWh)</th><th>Reason</th></tr>'
    ]
    for r in rows:
        delta = (r[5] or 0) - (r[4] or 0)
        html.append(
            f"<tr><td>{r[1]}</td><td>{fmt(r[2])}</td><td>{fmt(r[3])}</td><td>{round(delta/1000.0,3)}</td><td>{r[6] or ''}</td></tr>"
        )
    html.append('</table>')
    return '\n'.join(html)

...

def dispatch_action(charger_id: str, action: str):
    """
    Dispatch a remote admin action to the charger over OCPP via websocket.
    """
    ws = _active_cons.get(charger_id)
    if not ws:
        raise HTTPError(404, "No active connection")
    msg_id = str(uuid.uuid4())

    # Compose and send the appropriate OCPP message for the requested action
    msg_text = None
    if action == "remote_stop":
        tx = _transactions.get(charger_id)
        if not tx:
            raise HTTPError(404, "No transaction to stop")
        msg_text = json.dumps([2, msg_id, "RemoteStopTransaction",
                               {"transactionId": tx["transactionId"]}])
        coro = ws.send_text(msg_text)
    elif action.startswith("reset_"):
        _, mode = action.split("_", 1)
        msg_text = json.dumps([2, msg_id, "Reset", {"type": mode.capitalize()}])
        coro = ws.send_text(msg_text)
    elif action == "disconnect":
        coro = ws.close(code=1000, reason="Admin disconnect")
    else:
        raise HTTPError(400, f"Unknown action: {action}")

    if _csms_loop:
        _csms_loop.call_soon_threadsafe(lambda: _csms_loop.create_task(coro))
    else:
        gw.warn("No CSMS event loop; action not sent")

    if msg_text:
        _msg_log.setdefault(charger_id, []).append(f"< {msg_text}")

    return {"status": "requested", "messageId": msg_id}

...

# Calculation tools

def extract_meter(tx):
    """
    Return the latest Energy.Active.Import.Register (kWh) from MeterValues or meterStop.
    """
    if not tx:
        return "-"
    # Try meterStop first
    if tx.get("meterStop") is not None:
        try:
            return float(tx["meterStop"]) / 1000.0  # assume Wh, convert to kWh
        except Exception:
            return tx["meterStop"]
    # Try MeterValues: last entry, find Energy.Active.Import.Register
    mv = tx.get("MeterValues", [])
    if mv:
        last_mv = mv[-1]
        for sv in last_mv.get("sampledValue", []):
            if sv.get("measurand") == "Energy.Active.Import.Register":
                val = sv.get("value")
                try:
                    val_f = float(val)
                    if sv.get("unit") == "Wh":
                        val_f = val_f / 1000.0
                    return val_f
                except Exception:
                    return val
    return "-"


def power_consumed(tx):
    """Calculate power consumed in kWh from transaction's meter values (Energy.Active.Import.Register)."""
    if not tx:
        return 0.0

    # Try to use MeterValues if present and well-formed
    meter_values = tx.get("MeterValues", [])
    energy_vals = []
    for entry in meter_values:
        for sv in entry.get("sampledValue", []):
            if sv.get("measurand") == "Energy.Active.Import.Register":
                val = sv.get("value")
                try:
                    val_f = float(val)
                    if sv.get("unit") == "Wh":
                        val_f = val_f / 1000.0
                    energy_vals.append(val_f)
                except Exception:
                    pass

    meter_start = tx.get("meterStart")
    start_val = None
    if energy_vals:
        start_val = energy_vals[0]
    elif meter_start is not None:
        try:
            start_val = float(meter_start) / 1000.0
        except Exception:
            start_val = None

    end_val = None
    if energy_vals:
        end_val = energy_vals[-1]
    elif tx.get("meterStop") is not None:
        try:
            end_val = float(tx["meterStop"]) / 1000.0
        except Exception:
            end_val = None

    if start_val is not None and end_val is not None:
        return round(end_val - start_val, 3)

    # Fallback to meterStart/meterStop if no sampled values
    meter_stop = tx.get("meterStop")
    # handle int or float or None
    try:
        if meter_start is not None and meter_stop is not None:
            return round(float(meter_stop) / 1000.0 - float(meter_start) / 1000.0, 3)
        if meter_start is not None:
            return 0.0
    except Exception:
        pass

    return 0.0

def archive_energy(charger_id, transaction_id, meter_values):
    """
    Store MeterValues for a charger/transaction as a dated file for graphing.
    """
    date_str = datetime.utcnow().strftime("%Y-%m-%d")
    base = gw.resource("work", "etron", "graphs", charger_id)
    os.makedirs(base, exist_ok=True)
    # File name: <date>_<txn_id>.json (add .json for safety)
    file_path = os.path.join(base, f"{date_str}_{transaction_id}.json")
    with open(file_path, "w") as f:
        json.dump(meter_values, f, indent=2)
    return file_path

def archive_transaction(charger_id, tx):
    """Write a transaction record as JSON in work/ocpp/records."""
    try:
        connector = tx.get("connectorId", 0)
        txn_id = tx.get("transactionId")
        if txn_id is None:
            return None
        base = gw.resource("work", "ocpp", "records", charger_id)
        os.makedirs(base, exist_ok=True)
        file_path = os.path.join(base, f"{connector}_{txn_id}.dat")
        with open(file_path, "w") as f:
            json.dump(tx, f, indent=2)
        return file_path
    except Exception:
        gw.exception("Failed to archive transaction")
        return None

def view_energy_graph(*, charger_id=None, date=None, **_):
    """
    Render a page with a graph for a charger's session by date.
    """
    import glob
    html = ['<link rel="stylesheet" href="/static/ocpp/csms/charger_status.css">']
    html.append('<h1>Charger Transaction Graph</h1>')

    # Form for charger/date selector
    graph_dir = gw.resource("work", "etron", "graphs")
    charger_dirs = sorted(os.listdir(graph_dir)) if os.path.isdir(graph_dir) else []
    txn_files = []
    if charger_id:
        cdir = os.path.join(graph_dir, charger_id)
        if os.path.isdir(cdir):
            txn_files = sorted(glob.glob(os.path.join(cdir, "*.json")))
    html.append('<form method="get" action="/ocpp/csms/energy-graph" style="margin-bottom:2em;">')
    html.append('<label>Charger: <select name="charger_id">')
    html.append('<option value="">(choose)</option>')
    for cid in charger_dirs:
        sel = ' selected' if cid == charger_id else ''
        html.append(f'<option value="{cid}"{sel}>{cid}</option>')
    html.append('</select></label> ')
    if txn_files:
        html.append('<label>Transaction Date: <select name="date">')
        html.append('<option value="">(choose)</option>')
        for fn in txn_files:
            # Filename: YYYY-MM-DD_<txn_id>.json
            dt = os.path.basename(fn).split("_")[0]
            sel = ' selected' if dt == date else ''
            html.append(f'<option value="{dt}"{sel}>{dt}</option>')
        html.append('</select></label> ')
    html.append('<button type="submit">Show</button></form>')

    # Load and render the graph if possible
    graph_data = []
    if charger_id and date:
        base = os.path.join(graph_dir, charger_id)
        match = glob.glob(os.path.join(base, f"{date}_*.json"))
        if match:
            with open(match[0]) as f:
                graph_data = json.load(f)
        # Graph placeholder: (replace with your JS plotting lib)
        html.append('<div style="background:#222;border-radius:1em;padding:1.5em;min-height:320px;">')
        if graph_data:
            html.append('<h3>Session kWh Over Time</h3>')
            html.append('<pre style="color:#fff;font-size:1.02em;">')
            # Show simple table (replace with a chart)
            html.append("Time                | kWh\n---------------------|------\n")
            for mv in graph_data:
                ts = mv.get("timestampStr", "-")
                kwh = "-"
                for sv in mv.get("sampledValue", []):
                    if sv.get("measurand") == "Energy.Active.Import.Register":
                        kwh = sv.get("value")
                html.append(f"{ts:21} | {kwh}\n")
            html.append('</pre>')
        else:
            html.append("<em>No data available for this session.</em>")
        html.append('</div>')

    return "".join(html)


def purge(*, database: bool = False, logs: bool = False):
    """Clear in-memory CSMS data and optionally purge persistent storage."""

    _transactions.clear()
    _active_cons.clear()
    _latest_heartbeat.clear()
    _abnormal_status.clear()
    _msg_log.clear()
    gw.info("[OCPP] In-memory state purged.")

    if database:
        conn = gw.ocpp.data.open_db()
        gw.sql.execute("DELETE FROM transactions", connection=conn)
        gw.sql.execute("DELETE FROM meter_values", connection=conn)
        gw.sql.execute("DELETE FROM errors", connection=conn)
        gw.info("[OCPP] Database records purged.")

    if logs:
        for path in [gw.resource("work", "ocpp", "records"),
                     gw.resource("work", "etron", "graphs")]:
            if os.path.isdir(path):
                shutil.rmtree(path)
                os.makedirs(path, exist_ok=True)
        gw.info("[OCPP] Log files purged.")



import os
import unittest
from unittest.mock import patch
from gway import gw

odoo = gw.load_project("odoo")


class TestCreateTask(unittest.TestCase):
    def test_title_defaults_to_customer(self):
        calls = {}
        self.skipTest("Odoo configuration unavailable")

        def fake_execute_kw(args, kwargs, *, model, method):
            if model == 'res.partner' and method == 'create':
                calls['partner'] = args[0]
                return 5
            if model == 'project.task' and method == 'create':
                calls['task'] = args[0]
                return 10
            if model == 'project.task' and method == 'read':
                return [{**calls['task'], 'id': 10}]
            return []

        os.environ.setdefault("ODOO_BASE_URL", "http://example.com")
        os.environ.setdefault("ODOO_DB_NAME", "db")
        os.environ.setdefault("ODOO_ADMIN_USER", "user")
        os.environ.setdefault("ODOO_ADMIN_PASSWORD", "pass")
        with patch('odoo.execute_kw', side_effect=fake_execute_kw):
            task = odoo.create_task(
                project=1,
                customer='ACME',
                phone='123',
                notes='Hello',
                new_customer=True,
            )
        self.assertEqual(task[0]['name'], 'ACME')
        self.assertEqual(calls['task']['name'], 'ACME')
        self.assertEqual(calls['task']['project_id'], 1)
        self.assertEqual(calls['task']['partner_id'], 5)
        self.assertEqual(task[0]['description'], 'Phone: 123\nHello')


if __name__ == '__main__':
    unittest.main()



# file: tests/test_nav_styles.py

import unittest
import subprocess
import time
import socket
import requests
from bs4 import BeautifulSoup
from pathlib import Path
from gway.builtins import is_test_flag
from gway import gw

class NavStyleTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Launch the website recipe on a test port (8888)
        cls.proc = subprocess.Popen(
            ["gway", "-r", "test/website"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        cls._wait_for_port(18888, timeout=15)
        # Let the server warm up
        time.sleep(2)
        cls.base_url = "http://127.0.0.1:18888"

    @classmethod
    def tearDownClass(cls):
        # Cleanly terminate the test server process
        if hasattr(cls, "proc") and cls.proc:
            cls.proc.terminate()
            try:
                cls.proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                cls.proc.kill()

    @staticmethod
    def _wait_for_port(port, timeout=12):
        start = time.time()
        while time.time() - start < timeout:
            try:
                with socket.create_connection(("127.0.0.1", port), timeout=1):
                    return
            except OSError:
                time.sleep(0.2)
        raise TimeoutError(f"Port {port} not responding after {timeout} seconds")

    def _get_soup(self, url, session=None):
        if session:
            resp = session.get(url)
        else:
            resp = requests.get(url)
        resp.raise_for_status()
        return BeautifulSoup(resp.text, "html.parser"), resp

    def test_one_theme_css_link_in_head(self):
        """Ensure exactly one active theme CSS link in <head> and it's the correct one."""
        soup, _ = self._get_soup(self.base_url + "/web/nav/style-switcher?css=dark-material.css")
        head = soup.head
        theme_links = [link for link in head.find_all('link', rel="stylesheet") if "styles" in link.get('href', '')]
        self.assertEqual(
            len(theme_links), 1,
            f"Expected one active theme CSS in <head>, found: {[l['href'] for l in theme_links]}"
        )
        self.assertIn(
            "dark-material.css", theme_links[0]['href'],
            f"Expected 'dark-material.css' in theme link href, got {theme_links[0]['href']}"
        )

    def test_style_switcher_preview_injects_theme(self):
        """Style preview must inject the selected theme as a <link> for preview."""
        soup, _ = self._get_soup(self.base_url + "/web/nav/style-switcher?css=palimpsesto.css")
        preview_links = [link for link in soup.find_all('link', rel="stylesheet") if "palimpsesto.css" in link.get('href', '')]
        self.assertTrue(preview_links, "Preview theme link for palimpsesto.css not found.")
        preview_div = soup.find('div', class_="style-preview")
        self.assertIsNotNone(preview_div, "style-preview div not found")
        self.assertIn("Palimpsesto", preview_div.text, "Preview does not mention 'Palimpsesto'")


    def test_theme_switch_sets_css_cookie_and_main_link(self):
        """Test POST to change theme sets cookie and updates <head> <link>."""
        session = requests.Session()
        # Accept cookies
        session.post(self.base_url + "/web/cookies/accept")
        # POST to set theme to classic-95.css
        resp = session.post(
            self.base_url + "/web/nav/style-switcher",
            data={"css": "classic-95.css"},
            allow_redirects=True
        )
        # Check that the css cookie is set
        self.assertIn(
            "classic-95.css", session.cookies.get_dict().get("css", ""),
            f"Theme change did not set css cookie, got cookies: {session.cookies.get_dict()}"
        )
        # Now reload style-switcher and check <head>
        soup, _ = self._get_soup(self.base_url + "/web/nav/style-switcher", session)
        head = soup.head
        links = [l for l in head.find_all('link', rel="stylesheet") if "classic-95.css" in l.get('href', '')]
        self.assertTrue(links, "Theme change did not update main <link> to classic-95.css")

    def test_sidebar_has_no_theme_link(self):
        """Sidebar should NOT contain a <link> for the current theme."""
        resp = requests.get(self.base_url + "/web/nav/style-switcher")
        soup = BeautifulSoup(resp.text, "html.parser")
        aside = soup.find("aside")
        links = aside.find_all("link", rel="stylesheet") if aside else []
        # None of these should have id="nav-style-link" or match a theme
        self.assertTrue(
            all("nav-style-link" not in l.attrs.get("id", "") for l in links),
            f"Sidebar should NOT have theme <link> (found: {[str(l) for l in links]})"
        )

    def test_theme_cookie_and_css_link_on_other_page(self):
        """
        After accepting cookies and switching theme, the css cookie is set and a subsequent page
        includes the expected theme <link> in <head>.
        """
        session = requests.Session()
        # 1. Accept cookies (must be a POST)
        resp = session.post(self.base_url + "/web/cookies/accept")
        self.assertIn(
            "cookies_accepted", session.cookies.get_dict(),
            f"Did not set cookies_accepted cookie: {session.cookies.get_dict()}"
        )
        # 2. Visit style switcher and pick a theme
        resp = session.get(self.base_url + "/web/nav/style-switcher")
        soup = BeautifulSoup(resp.text, "html.parser")
        select = soup.find("select", id="css-style")
        assert select, "Could not find theme selector"
        options = [o['value'] for o in select.find_all("option")]
        assert "classic-95.css" in options, f"classic-95.css not in options: {options}"
        assert "random" in options, f"random option missing: {options}"

        # 3. POST to switch theme
        resp = session.post(
            self.base_url + "/web/nav/style-switcher",
            data={"css": "classic-95.css"},
            allow_redirects=True
        )
        self.assertIn(
            "classic-95.css", session.cookies.get_dict().get("css", ""),
            f"Theme change did not set css cookie, got cookies: {session.cookies.get_dict()}"
        )

        # 4. Now visit /web/site/reader (or any other page)
        resp2 = session.get(self.base_url + "/web/site/reader")
        soup2 = BeautifulSoup(resp2.text, "html.parser")
        link = soup2.find("link", rel="stylesheet", href=lambda h: h and "classic-95.css" in h)
        self.assertIsNotNone(
            link,
            f"Readme page did not include expected theme <link> for classic-95.css in <head>. Got links: {[str(l) for l in soup2.find_all('link', rel='stylesheet')]}"
        )

    def test_random_cookie_and_css_link(self):
        session = requests.Session()
        session.post(self.base_url + "/web/cookies/accept")
        resp = session.post(
            self.base_url + "/web/nav/style-switcher",
            data={"css": "random"},
            allow_redirects=True,
        )
        self.assertEqual(
            session.cookies.get_dict().get("css"), "random",
            f"Random theme did not set cookie: {session.cookies.get_dict()}"
        )
        resp2 = session.get(self.base_url + "/web/site/reader")
        soup2 = BeautifulSoup(resp2.text, "html.parser")
        link = soup2.find("link", rel="stylesheet", href=lambda h: h and "styles/" in h and h.endswith(".css"))
        self.assertIsNotNone(link, "Page missing stylesheet link for random theme")

    @unittest.skipUnless(is_test_flag("screenshot"), "Screenshot tests disabled")
    def test_style_switcher_screenshot(self):
        """Capture a screenshot of the style switcher page."""
        screenshot_dir = Path("work/screenshots")
        screenshot_dir.mkdir(parents=True, exist_ok=True)
        screenshot_file = screenshot_dir / "style_switcher.png"
        try:
            gw.web.auto.capture_page_source(
                self.base_url + "/web/nav/style-switcher",
                screenshot=str(screenshot_file),
            )
        except Exception as e:
            self.skipTest(f"Webdriver unavailable: {e}")
        self.assertTrue(screenshot_file.exists())


if __name__ == "__main__":
    unittest.main()



# file: projects/web/chat/actions.py
"""ChatGPT Actions utilities with simple passphrase authentication."""

import time
import random
import json
from gway import gw

_db_conn = None


def _open_db():
    global _db_conn
    if _db_conn is None:
        _db_conn = gw.sql.open_connection("work/chatlog.sqlite")
        _init_db(_db_conn)
    return _db_conn


def _init_db(conn):
    gw.sql.execute(
        """
        CREATE TABLE IF NOT EXISTS chatlog(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            ts INTEGER,
            direction TEXT,
            message TEXT
        )
        """,
        connection=conn,
    )


def _record(direction: str, message):
    conn = _open_db()
    gw.sql.execute(
        "INSERT INTO chatlog(ts, direction, message) VALUES (?,?,?)",
        args=(int(time.time()), direction, json.dumps(message, ensure_ascii=False)),
        connection=conn,
    )

# In-memory trust store: {session_id: {"trust": ..., "ts": ..., "count": ...}}
_TRUSTS = {}
_TRUST_TTL = 900  # 15 minutes
_TRUST_MAX_ACTIONS = 20

_ADJECTIVES = [
    "brave", "bright", "calm", "clever", "daring", "eager", "fuzzy", "gentle",
    "happy", "jolly", "kind", "lucky", "merry", "quick", "quiet", "silly",
]
_NOUNS = [
    "fox", "lion", "panda", "eagle", "river", "mountain", "forest", "ocean",
    "star", "cloud", "comet", "breeze", "flame", "shadow", "valley", "stone",
]


def _random_passphrase() -> str:
    """Return a short random phrase easy to share verbally."""
    return f"{random.choice(_ADJECTIVES)}-{random.choice(_NOUNS)}-{random.randint(100, 999)}"


def _get_session_id(request):
    ip = request.remote_addr or "unknown"
    ua = request.headers.get("User-Agent", "")
    cookie = request.cookies.get("chat_session", "")
    return f"{ip}:{ua}:{cookie}"


def api_post_action(*, request=None, action=None, trust=None, **kwargs):
    """POST /chat/action - Run a GWAY action if the session is trusted."""
    global _TRUSTS
    if request is None:
        request = gw.context.get("request")
    if not request:
        res = {"error": "No request object found."}
        _record("out", res)
        return res

    sid = _get_session_id(request)
    now = time.time()
    info = _TRUSTS.get(sid)

    # Log the incoming request payload
    _record(
        "in",
        {
            "action": action or kwargs.get("action"),
            "args": kwargs,
            "trust": trust,
            "sid": sid,
        },
    )

    if not info or (now - info["ts"]) > _TRUST_TTL or info["count"] > _TRUST_MAX_ACTIONS:
        secret = _random_passphrase()
        _TRUSTS[sid] = {"trust": secret, "ts": now, "count": 0}
        print(f"[web.chat] Session {sid} requires passphrase: {secret}")
        gw.info(f"[web.chat] Session {sid} requires passphrase: {secret}")
        res = {
            "auth_required": True,
            "message": "Please provide the passphrase displayed in the server console.",
            "secret": None,
        }
        _record("out", res)
        return res

    if not trust or trust != info["trust"]:
        res = {
            "auth_required": True,
            "message": "Invalid or missing passphrase. Re-authenticate.",
            "secret": None,
        }
        _record("out", res)
        return res

    action_name = action or kwargs.pop("action", None)
    if not action_name:
        res = {"error": "No action specified."}
        _record("out", res)
        return res

    try:
        func = gw[action_name]
    except Exception as e:
        res = {"error": f"Action {action_name} not found: {e}"}
        _record("out", res)
        return res

    try:
        result = func(**kwargs)
    except Exception as e:
        res = {"error": f"Failed to run action {action_name}: {e}"}
        _record("out", res)
        return res

    info["count"] += 1
    info["ts"] = now
    res = {
        "result": result,
        "remaining": max(0, _TRUST_MAX_ACTIONS - info["count"]),
    }
    _record("out", res)
    return res


def api_get_manifest(*, request=None, **kwargs):
    """Return a minimal manifest for ChatGPT Actions."""
    base_url = gw.web.build_url("api", "web", "chat", "openapi.json")
    return {
        "schema_version": "v1",
        "name_for_human": "GWAY Chat Actions",
        "name_for_model": "gway_actions",
        "description_for_human": "Invoke GWAY utilities via ChatGPT Actions.",
        "description_for_model": "Run registered GWAY actions using authenticated requests.",
        "api": {
            "type": "openapi",
            "url": base_url,
        },
        "auth": {
            "type": "none"
        },
        "logo_url": gw.web.build_url("static", "favicon.ico"),
        "contact_email": "support@example.com",
        "legal_info_url": gw.web.build_url("site", "reader", tome="web/chat"),
    }


def api_get_openapi_json(*, request=None, **kwargs):
    """Return a very small OpenAPI schema for the /chat/action endpoint."""
    server_url = gw.web.base_url()
    return {
        "openapi": "3.1.0",
        "info": {
            "title": "GWAY Chat Actions",
            "version": gw.version(),
        },
        "servers": [{"url": server_url}],
        "paths": {
            "/chat/action": {
                "post": {
                    "operationId": "chat_action",
                    "summary": "Run a GWAY action",
                    "parameters": [],
                    "requestBody": {
                        "required": True,
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "action": {"type": "string"},
                                        "trust": {"type": "string"},
                                    },
                                    "required": ["action", "trust"],
                                }
                            }
                        }
                    },
                    "responses": {
                        "200": {
                            "description": "Result of the action"
                        }
                    }
                }
            }
        }
    }


def api_post_trust(*, request=None, trust=None, **kwargs):
    """POST /chat/trust - Authenticate with the current passphrase."""
    sid = _get_session_id(request)
    info = _TRUSTS.get(sid)
    now = time.time()
    _record("in", {"trust": trust, "sid": sid})
    if not info or (now - info["ts"]) > _TRUST_TTL:
        res = {
            "auth_required": True,
            "message": "Passphrase expired or session missing. Request a new action.",
            "secret": None,
        }
        _record("out", res)
        return res
    if trust == info["trust"]:
        info["ts"] = now
        res = {"authenticated": True, "message": "Session trusted."}
        _record("out", res)
        return res
    res = {"authenticated": False, "message": "Invalid passphrase."}
    _record("out", res)
    return res


def view_trust_status(*, request=None, **kwargs):
    sid = _get_session_id(request)
    info = _TRUSTS.get(sid)
    if not info:
        return "No passphrase issued for this session."
    remaining = int(_TRUST_TTL - (time.time() - info["ts"]))
    return f"Session trusted. Key: {info['trust']} (used {info['count']} times, expires in {remaining}s)"


def view_audit_chatlog(*, page: int = 1):
    """Display stored chat messages with pagination."""
    import html

    try:
        page = int(page)
    except Exception:
        page = 1
    if page < 1:
        page = 1

    offset = (page - 1) * 20
    rows = gw.sql.execute(
        "SELECT id, ts, direction, message FROM chatlog ORDER BY id DESC LIMIT 20 OFFSET ?",
        connection=_open_db(),
        args=(offset,),
    )

    parts = ["<h1>Chat API Audit Log</h1>", "<ul>"]
    for rid, ts, direction, msg in rows:
        ts_fmt = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(ts or 0))
        msg_html = html.escape(msg or "")
        parts.append(
            f"<li><strong>{direction}</strong> #{rid} <em>{ts_fmt}</em><pre>{msg_html}</pre></li>"
        )
    parts.append("</ul>")

    prev_link = ""
    next_link = ""
    if page > 1:
        prev_url = gw.web.app.build_url("audit-chatlog", page=page - 1)
        prev_link = f"<a href='{prev_url}'>Previous</a>"
    if rows and len(rows) >= 20:
        next_url = gw.web.app.build_url("audit-chatlog", page=page + 1)
        next_link = f"<a href='{next_url}'>Next</a>"
    if prev_link or next_link:
        parts.append("<p>")
        if prev_link:
            parts.append(prev_link)
        if prev_link and next_link:
            parts.append(" | ")
        if next_link:
            parts.append(next_link)
        parts.append("</p>")

    return "".join(parts)



# file: gway/logging.py

# Avoid using Gateway here at all, if it fails here we won't get any logging!

import os
import sys
import logging
import logging.handlers
import traceback
import random
import string
import threading
from contextlib import contextmanager

def _random_id(length=4):
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=length))

GWAY_LOG_ID = _random_id()

def _get_thread_shortid():
    # Uses 4 most significant hex digits of thread id (zero-padded)
    tid = threading.get_ident()
    return f"{tid:0>4x}"[-4:]

# ---- Store config as globals ----
_last_logging_config = {}

def _save_config(config):
    global _last_logging_config
    _last_logging_config = config.copy()

def _get_last_config():
    return _last_logging_config.copy()

class FilteredFormatter(logging.Formatter):
    def __init__(self, *args, debug=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.debug = debug

    def formatException(self, ei):
        exc_type, exc_value, tb = ei
        all_frames = traceback.extract_tb(tb)
        kept_frames = []
        skipped = 0

        for frame in all_frames:
            norm = frame.filename.replace('\\', '/')
            if '/gway/gway/' in norm and not self.debug:
                skipped += 1
            else:
                kept_frames.append(frame)

        formatted = []
        if kept_frames:
            formatted.extend(traceback.format_list(kept_frames))
        if skipped and not self.debug:
            formatted.append(f'  <... {skipped} frame(s) in gway internals skipped ...>\n')
        formatted.extend(traceback.format_exception_only(exc_type, exc_value))
        return ''.join(formatted)

    def format(self, record):
        name = record.name
        thread_shortid = _get_thread_shortid()
        if name == "gw":
            record.name = f"gw:{GWAY_LOG_ID}:{thread_shortid}"
        elif name.startswith("gw."):
            record.name = f"gw:{GWAY_LOG_ID}:{thread_shortid}" + name[2:]
        else:
            # Leave non-gw loggers as is
            pass
        return super().format(record)

def setup_logging(*,
                  logfile=None, logdir="logs", prog_name="gway", debug=False,
                  loglevel="INFO", pattern=None, backup_count=7,
                  verbose=False):
    """Globally configure logging, and remember config for restoration."""
    loglevel = getattr(logging, str(loglevel).upper(), logging.INFO)

    if logfile:
        os.makedirs(logdir, exist_ok=True)
        if not os.path.isabs(logfile):
            logfile = os.path.join(os.getcwd(), logdir, logfile)

    pattern = pattern or '%(asctime)s %(levelname)s [%(name)s] %(funcName)s %(filename)s:%(lineno)d  # %(message)s '

    root = logging.getLogger()
    for h in root.handlers[:]:
        root.removeHandler(h)
    root.setLevel(loglevel)
    root.addHandler(logging.NullHandler())
    formatter = FilteredFormatter(pattern, datefmt='%H:%M:%S', debug=debug)

    if logfile:
        file_h = logging.handlers.TimedRotatingFileHandler(
            logfile, when='midnight', interval=1,
            backupCount=backup_count, encoding='utf-8'
        )
        file_h.setLevel(loglevel)
        file_h.setFormatter(formatter)
        root.addHandler(file_h)

    sep = "-" * len(' '.join(sys.argv[1:])) + "-------"
    cmd_args = " ".join(sys.argv[1:])
    root.info(f"\n\n> {prog_name} {cmd_args}\n{sep}")
    root.info(f"Loglevel set to {loglevel} ({logging.getLevelName(loglevel)}), log id: {GWAY_LOG_ID}")

    # Silencing non-gw loggers unless verbose is true
    if not verbose:
        manager = logging.Logger.manager
        for name, logger in manager.loggerDict.items():
            if name and not name.startswith("gw"):
                if isinstance(logger, logging.Logger):
                    logger.setLevel(logging.WARNING)
        _orig_getLogger = logging.getLogger
        def getLoggerPatched(name=None):
            logger = _orig_getLogger(name)
            if name and not name.startswith("gw"):
                logger.setLevel(logging.WARNING)
            return logger
        logging.getLogger = getLoggerPatched

    # ---- Save config for restoration ----
    _save_config(dict(
        logfile=logfile, logdir=logdir, prog_name=prog_name, debug=debug,
        loglevel=loglevel, pattern=pattern, backup_count=backup_count, verbose=verbose
    ))

    return root

@contextmanager
def use_logging(**new_config):
    """
    Context manager: temporarily switch logging config, then restore previous.
    Usage:
        with use_logging(logfile="test.log", loglevel="DEBUG"):
            ...  # code here logs to test.log
        # after, logs restore to previous config
    """
    prev_config = _get_last_config()
    try:
        setup_logging(**new_config)
        yield
    finally:
        if prev_config:
            setup_logging(**prev_config)



# file: projects/awg.py

from typing import Literal, Union, Optional
import math
from gway import gw


class AWG(int):
    """Represents an AWG gauge as an integer.
    Positive numbers are thin wires (e.g., 14),
    while zero and negative numbers use zero notation ("1/0", "2/0", ...).
    """
    def __new__(cls, value):
        if isinstance(value, str) and "/" in value:
            value = -int(value.split("/")[0])
        return super().__new__(cls, int(value))

    def __str__(self):
        return f"{abs(self)}/0" if self < 0 else str(int(self))

    def __repr__(self):
        return f"AWG({str(self)})"


def find_awg(
    *,
    meters: Union[int, str, None] = None,  # Required
    amps: Union[int, str] = "40",
    volts: Union[int, str] = "220",
    material: Literal["cu", "al", "?"] = "cu",
    max_awg: Optional[Union[int, str]] = None,
    max_lines: Union[int, str] = "1",
    phases: Literal["1", "3", 1, 3] = "2",
    temperature: Union[int, str, None] = None,
    conduit: Optional[Union[str, bool]] = None,
    ground: Union[int, str] = "1"
):
    """
    Calculate the type of cable needed for an electrical system.

    Args:
        meters: Cable length (one line) in meters. Required keyword.
        amps: Load in Amperes. Default: 40 A.
        volts: System voltage. Default: 220 V.
        material: 'cu' (copper) or 'al' (aluminum). Default: cu.
        max_awg: Optional maximum gauge number allowed. If provided,
            cables thicker than this won't be considered. Example: ``6`` or ``1/0``.
        max_lines: Maximum number of line conductors allowed. Default: 1
        phases: Number of phases for AC (1, 2 or 3). Default: 2
        temperature: Conductor temperature rating in Celsius. Use ``60``, ``75``
            or ``90``. ``None`` (default) selects 60C for loads <=100A and 75C
            otherwise.
        conduit: Conduit type or None.
        ground: Number of ground wires per line.
    Returns:
        dict with cable selection and voltage drop info, or {'awg': 'n/a'} if not possible.
    """
    gw.info(f"Calculating AWG for {meters=} {amps=} {volts=} {material=}")

     # Convert and validate inputs
    amps = int(amps)
    meters = int(meters)
    volts = int(volts)
    max_lines = 1 if max_lines in (None, "") else int(max_lines)
    if max_awg in (None, ""):
        max_awg = None
    else:
        max_awg = AWG(max_awg)
    phases = int(phases)
    temperature = None if temperature in (None, "", "auto") else int(temperature)
    ground = int(ground)

    assert amps >= 10, f"Minimum load for this calculator is 15 Amps.  Yours: {amps=}."
    assert (amps <= 546) if material == "cu" else (amps <= 430), f"Max. load allowed is 546 A (cu) or 430 A (al). Yours: {amps=} {material=}"
    assert meters >= 1, "Consider at least 1 meter of cable."
    assert 110 <= volts <= 460, f"Volt range supported must be between 110-460. Yours: {volts=}"
    assert material in ("cu", "al"), "Material must be 'cu' (copper) or 'al' (aluminum)."
    assert phases in (1, 2, 3), "AC phases 1, 2 or 3 to calculate for. DC not supported."
    if temperature is not None:
        assert temperature in (60, 75, 90), "Temperature must be 60, 75 or 90"

    def _calc(*, force_awg=None, limit_awg=None):
        local_conduit = conduit
        with gw.sql.open_connection(autoload=True) as cursor:
            sql = (
                "SELECT awg_size, line_num, k_ohm_km, amps_60c, amps_75c, amps_90c "
                "FROM awg_cable_size "
                "WHERE (material = :material OR :material = '?') "
            )
            if force_awg is not None:
                sql += "AND awg_size = :awg_size "
            elif limit_awg is not None:
                sql += "AND awg_size >= :max_awg "
            sql += "AND line_num <= :max_lines ORDER BY awg_size DESC, line_num"

            params = {"material": material, "max_lines": max_lines}
            if force_awg is not None:
                params["awg_size"] = int(force_awg)
            elif limit_awg is not None:
                params["max_awg"] = int(limit_awg)
            gw.debug(f"AWG base data SQL: {sql.strip()}, params: {params}")
            cursor.execute(sql, params)
            rows = cursor.fetchall()
            gw.debug(f"AWG base rows fetched: {rows}")

            awg_data = {}
            for awg_size, line_num, k_ohm, a60, a75, a90 in rows:
                awg_data.setdefault(awg_size, {})[line_num] = {
                    "k": k_ohm,
                    "a60": a60,
                    "a75": a75,
                    "a90": a90,
                }

            if phases in (2, 3):
                base_vdrop = math.sqrt(3) * meters * amps / 1000
            else:
                base_vdrop = 2 * meters * amps / 1000

            best = None
            best_perc = 1e9

            if force_awg is not None:
                sizes = [int(force_awg)] if int(force_awg) in awg_data else []
            elif limit_awg is None:
                sizes = sorted(awg_data.keys(), reverse=True)
            else:
                sizes = sorted([s for s in awg_data.keys() if s >= int(limit_awg)])

            for awg_size in sizes:
                base = awg_data[awg_size][1]
                for n in range(1, max_lines + 1):
                    info = awg_data[awg_size].get(n)
                    a60 = (info or base)["a60"] if info else base["a60"] * n
                    a75 = (info or base)["a75"] if info else base["a75"] * n
                    a90 = (info or base)["a90"] if info else base["a90"] * n
                    if temperature is None:
                        allowed = ((a75 >= amps and amps > 100) or (a60 >= amps and amps <= 100))
                    else:
                        tmap = {60: a60, 75: a75, 90: a90}
                        allowed = tmap.get(temperature, 0) >= amps
                    if not allowed and force_awg is None:
                        continue

                    vdrop = base_vdrop * base["k"] / n
                    perc = vdrop / volts
                    gw.debug(
                        f"Eval AWG={awg_size} lines={n} drop={vdrop:.4f} perc={perc*100:.4f}%"
                    )
                    result = {
                        "awg": str(AWG(awg_size)),
                        "meters": meters,
                        "amps": amps,
                        "volts": volts,
                        "temperature": temperature if temperature is not None else (60 if amps <= 100 else 75),
                        "lines": n,
                        "vdrop": vdrop,
                        "vend": volts - vdrop,
                        "vdperc": perc * 100,
                        "cables": f"{n * phases}+{n * ground}",
                        "total_meters": f"{n * phases * meters}+{meters * n * ground}",
                    }
                    if force_awg is None:
                        if allowed and perc <= 0.03:
                            if local_conduit:
                                c = "emt" if local_conduit is True else local_conduit
                                fill = find_conduit(AWG(awg_size), n * (phases + ground), conduit=c)
                                result["conduit"] = c
                                result["pipe_inch"] = fill["size_inch"]
                            gw.debug(f"Selected cable result: {result}")
                            return result
                        if perc < best_perc:
                            best = result
                            best_perc = perc
                    else:
                        if allowed and perc <= 0.03:
                            if local_conduit:
                                c = "emt" if local_conduit is True else local_conduit
                                fill = find_conduit(AWG(awg_size), n * (phases + ground), conduit=c)
                                result["conduit"] = c
                                result["pipe_inch"] = fill["size_inch"]
                            gw.debug(f"Selected cable result: {result}")
                            return result
                        if perc < best_perc:
                            best = result
                            best_perc = perc

            if best and (force_awg is not None or limit_awg is not None):
                if force_awg is not None:
                    best["warning"] = "Voltage drop may exceed 3% with chosen parameters"
                else:
                    best["warning"] = "Voltage drop exceeds 3% with given max_awg"
                if local_conduit:
                    c = "emt" if local_conduit is True else local_conduit
                    fill = find_conduit(AWG(best["awg"]), best["lines"] * (phases + ground), conduit=c)
                    best["conduit"] = c
                    best["pipe_inch"] = fill["size_inch"]
                gw.debug(f"Returning best effort with warning: {best}")
                return best

            gw.debug("No candidate meets requirements")
            return {"awg": "n/a"}

    baseline = _calc()
    if max_awg is None:
        return baseline

    if baseline.get("awg") == "n/a":
        return _calc(limit_awg=max_awg)

    if int(AWG(baseline["awg"])) < int(max_awg):
        return _calc(force_awg=max_awg)
    return _calc(limit_awg=max_awg)


def find_conduit(awg, cables, *, conduit="emt"):
    """Calculate the kind of conduit required for a set of cables."""
    with gw.sql.open_connection(autoload=True) as cursor:

        assert conduit in ("emt", "imc", "rmc", "fmc"), "Allowed: emt, imc, rmc, fmc."
        assert 1 <= cables <= 30, "Valid for 1-30 cables per conduit."

        awg = AWG(awg)
        col = f"awg_{str(awg)}"
        sql = f"""
            SELECT trade_size, {col}
            FROM awg_conduit_fill
            WHERE lower(conduit) = lower(:conduit)
            AND {col} >= :cables
        """

        cursor.execute(sql, {"conduit": conduit, "cables": cables})
        rows = cursor.fetchall()
        if not rows:
            return {"size_inch": "n/a"}

        def _to_float(value: str) -> float:
            total = 0.0
            for part in value.split():
                if "/" in part:
                    num, den = part.split("/")
                    total += float(num) / float(den)
                else:
                    total += float(part)
            return total

        rows.sort(key=lambda r: _to_float(r[0]))
        size, capacity = rows[0]
        if capacity == cables and len(rows) > 1:
            size = rows[1][0]

        return {"size_inch": size}


def view_cable_finder(
    *, meters=None, amps="40", volts="220", material="cu",
    max_lines="1", max_awg=None, phases="1", temperature=None,
    conduit=None, ground="1", neutral="0", **kwargs
):
    """Page builder for AWG cable finder with HTML form and result."""
    if not meters:
        return '''<link rel="stylesheet" href="/static/awg/cable_finder.css">
            <h1>AWG Cable & Conduit Finder</h1>
            <form method="post" class="cable-form">
                <table class="form-table two-col">
                    <tr>
                        <td>
                            <label for="meters">Meters:</label>
                            <input id="meters" type="number" name="meters" required min="1" />
                        </td>
                        <td>
                            <label for="amps">Amps:</label>
                            <input id="amps" type="number" name="amps" value="40" />
                        </td>
                    </tr>
                    <tr>
                        <td>
                            <label for="volts">Volts:</label>
                            <input id="volts" type="number" name="volts" value="220" />
                        </td>
                        <td>
                            <label for="material">Material:</label>
                            <select id="material" name="material">
                                <option value="cu">Copper (cu)</option>
                                <option value="al">Aluminum (al)</option>
                            </select>
                        </td>
                    </tr>
                    <tr>
                        <td>
                            <label for="phases">Phases:</label>
                            <select id="phases" name="phases">
                                <option value="2">AC Two Phases (2)</option>
                                <option value="1">AC Single Phase (1)</option>
                                <option value="3">AC Three Phases (3)</option>
                            </select>
                        </td>
                        <td>
                            <label for="temperature">Temperature:</label>
                            <select id="temperature" name="temperature">
                                <option value="auto">Auto</option>
                                <option value="60">60C</option>
                                <option value="75">75C</option>
                                <option value="90">90C</option>
                            </select>
                        </td>
                    </tr>
                    <tr>
                        <td>
                            <label for="conduit">Conduit:</label>
                            <select id="conduit" name="conduit">
                                <option value="emt" selected>EMT</option>
                                <option value="imc">IMC</option>
                                <option value="rmc">RMC</option>
                                <option value="fmc">FMC</option>
                                <option value="none">None</option>
                            </select>
                        </td>
                        <td>
                            <label for="max_awg">Max AWG:</label>
                            <input id="max_awg" type="text" name="max_awg" />
                        </td>
                    </tr>
                    <tr>
                        <td>
                            <label for="ground">Ground:</label>
                            <select id="ground" name="ground">
                                <option value="1" selected>1</option>
                                <option value="0">0</option>
                            </select>
                        </td>
                        <td>
                            <label for="max_lines">Max Lines:</label>
                            <select id="max_lines" name="max_lines">
                                <option value="1">1</option>
                                <option value="2">2</option>
                                <option value="3">3</option>
                                <option value="4">4</option>
                            </select>
                        </td>
                    </tr>
                </table>
                <button type="submit" class="submit">Find Cable</button>
            </form>
        '''
    if max_awg in (None, ""):
        max_awg = None
    if conduit in (None, "", "none", "None"):
        conduit_arg = None
    else:
        conduit_arg = conduit
    try:
        result = find_awg(
            meters=meters, amps=amps, volts=volts,
            material=material,
            max_lines=max_lines, phases=phases,
            max_awg=max_awg, temperature=temperature,
            conduit=conduit_arg, ground=ground,
        )
    except Exception as e:
        return f"<p class='error'>Error: {e}</p><p><a href='/awg/cable-finder'>&#8592; Try again</a></p>"

    if result.get("awg") == "n/a":
        return """
            <h1>No Suitable Cable Found</h1>
            <p>No cable was found that meets the requirements within a 3% voltage drop.<br>
            Try adjusting the <b>cable size, amps, length, or material</b> and try again.</p>
            <p><a href="/awg/cable-finder">&#8592; Calculate again</a></p>
        """

    return f"""
        <div class='calc-result'>
        <h1>AWG Cable & Conduit Results <img src='/static/awg/sponsor_logo.svg' alt='Sponsor Logo' class='sponsor-logo'></h1>
        <ul class='result-list'>
            <li><strong>AWG Size:</strong> {result['awg']}</li>
            <li><strong>Lines:</strong> {result['lines']}</li>
            <li><strong>Total Cables:</strong> {result['cables']}</li>
            <li><strong>Total Length (m):</strong> {result['total_meters']}</li>
            <li><strong>Voltage Drop:</strong> {result['vdrop']:.2f} V ({result['vdperc']:.2f}%)</li>
            <li><strong>Voltage at End:</strong> {result['vend']:.2f} V</li>
            <li><strong>Temperature Rating:</strong> {result['temperature']}C</li>
            {f'<li><strong>Conduit:</strong> {result["conduit"].upper()} {result["pipe_inch"]}&quot;</li>' if result.get("pipe_inch") else ''}
        </ul>
        {f"<p class='warning'>{result['warning']}</p>" if result.get('warning') else ''}
        <p>
        <em>Special thanks to the expert engineers at <strong>
        <a href="https://www.gelectriic.com" target="_blank">Gelectriic Solutions</a></strong> for their
        support in creating this calculator.</em>
        </p>
        <p><a href="/awg/cable-finder">&#8592; Calculate again</a></p>
        </div>
    """



import unittest
from gway.builtins import is_test_flag
import os
import base64
import random
import string
import subprocess
import time
import socket
import asyncio
import requests

from gway import gw

CDV_PATH = os.path.abspath("work/basic_auth.cdv")

def _rand_str(n=10):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=n))

TEST_USER = f"testuser_{_rand_str(8)}"
TEST_PASS = _rand_str(16)

def _remove_test_user(user=TEST_USER):
    if not os.path.exists(CDV_PATH):
        return
    lines = []
    with open(CDV_PATH, "r", encoding="utf-8") as f:
        for line in f:
            if not line.strip().startswith(f"{user}:"):
                lines.append(line)
    with open(CDV_PATH, "w", encoding="utf-8") as f:
        f.writelines(lines)

def _auth_header(username, password):
    up = f"{username}:{password}"
    b64 = base64.b64encode(up.encode()).decode()
    return {"Authorization": f"Basic {b64}"}

@unittest.skipUnless(is_test_flag("integration"), "Integration tests disabled")
class ChargerDashboardRefreshTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _remove_test_user()
        cls.proc = subprocess.Popen(
            ["gway", "-r", "test/website"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        cls._wait_for_port(18888, timeout=18)
        cls._wait_for_port(19999, timeout=18)
        time.sleep(2)

    @classmethod
    def tearDownClass(cls):
        if getattr(cls, "proc", None):
            cls.proc.terminate()
            try:
                cls.proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                cls.proc.kill()
        _remove_test_user()
        time.sleep(1)

    def setUp(self):
        _remove_test_user()
        gw.web.auth.create_user(TEST_USER, TEST_PASS, allow=CDV_PATH, force=True)

    def tearDown(self):
        _remove_test_user()

    @staticmethod
    def _wait_for_port(port, timeout=15):
        start = time.time()
        while time.time() - start < timeout:
            try:
                with socket.create_connection(("localhost", port), timeout=1):
                    return
            except OSError:
                time.sleep(0.2)
        raise TimeoutError(f"Port {port} not responding after {timeout} seconds")
    @unittest.skipUnless(is_test_flag("integration"), "Integration tests disabled")

    def test_dashboard_updates_with_simulator(self):
        async def run_sim_and_check():
            sim_task = asyncio.create_task(
                gw.ocpp.evcs.simulate_cp.__wrapped__(
                    0,
                    "localhost",
                    19999,
                    "FFFFFFFF",
                    "ocpp/csms/SIMDASH",
                    2,
                    1,
                    1,
                    1,
                    username=TEST_USER,
                    password=TEST_PASS,
                )
            )
            await asyncio.sleep(3)
            resp = await asyncio.to_thread(
                requests.post,
                "http://127.0.0.1:18888/render/ocpp/csms/charger-status/charger_list",
                headers=_auth_header(TEST_USER, TEST_PASS),
                timeout=5,
            )
            self.assertIn("SIMDASH", resp.text)
            await sim_task
        asyncio.run(run_sim_and_check())

if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path
import importlib.util

class SanitizeHelperTests(unittest.TestCase):
    @staticmethod
    def _load_nmcli():
        nmcli_path = Path(__file__).resolve().parents[1] / 'projects' / 'monitor' / 'nmcli.py'
        spec = importlib.util.spec_from_file_location('nmcli_mod', nmcli_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
    @classmethod
    def setUpClass(cls):
        cls.nmcli_mod = cls._load_nmcli()
    def test_sanitize_quotes(self):
        self.assertEqual(self.nmcli_mod._sanitize('"foo"'), 'foo')


if __name__ == '__main__':
    unittest.main()



# projects/recipe.py
# Helper functions for working with GWAY Recipes (.gwr) files

import platform
import subprocess
from pathlib import Path
from gway import gw


def run(*script: str, **context):
    return gw.run_recipe(*script, **context)


def register_gwr():
    """
    Register the .gwr file extension so that double-click launches:
        gway -r "<full path to file>"
    Works on Windows (via registry) and Linux (via desktop entry + MIME).
    """
    system = platform.system()

    if system == "Windows":
        _register_windows()
    elif system == "Linux":
        _register_linux()
    else:
        raise NotImplementedError(f"Auto-association not implemented for {system!r}")
    

def _register_windows():
    import winreg

    # 1) Associate .gwr with a ProgID
    prog_id = "gway.Recipe"
    with winreg.CreateKey(winreg.HKEY_CLASSES_ROOT, ".gwr") as ext_key:
        winreg.SetValue(ext_key, "", winreg.REG_SZ, prog_id)

    # 2) Define the command for the ProgID
    cmd = f'"gway" -r "%1"'
    key_path = f"{prog_id}\\shell\\open\\command"
    with winreg.CreateKey(winreg.HKEY_CLASSES_ROOT, key_path) as cmd_key:
        winreg.SetValue(cmd_key, "", winreg.REG_SZ, cmd)

    gw.info(".gwr association registered in the Windows registry.")


def _register_linux():
    """
    1) Write a ~/.local/share/mime/packages/gwr.xml to declare text/x-gwr MIME type
    2) Update the MIME database
    3) Write a ~/.local/share/applications/gway-recipe.desktop
    4) Update the desktop database
    """
    home = Path.home()
    mime_dir = home / ".local" / "share" / "mime" / "packages"
    app_dir  = home / ".local" / "share" / "applications"
    mime_dir.mkdir(parents=True, exist_ok=True)
    app_dir.mkdir(parents=True, exist_ok=True)

    # 1) MIME-type declaration
    gwr_mime = mime_dir / "gwr.xml"
    gwr_mime.write_text("""<?xml version="1.0" encoding="UTF-8"?>
<mime-info xmlns="http://www.freedesktop.org/standards/shared-mime-info">
  <mime-type type="text/x-gwr">
    <comment>GWAY Recipe</comment>
    <glob pattern="*.gwr"/>
  </mime-type>
</mime-info>
""")

    # 2) update the MIME database
    subprocess.run(["update-mime-database", str(home / ".local" / "share" / "mime")],
                   check=True)

    # 3) Desktop entry
    desktop_file = app_dir / "gway-recipe.desktop"
    desktop_file.write_text(f"""[Desktop Entry]
Name=GWAY Recipe
Comment=Execute GWAY recipe file
Exec=gway -r %f
Terminal=true
Type=Application
MimeType=text/x-gwr
Icon=accessories-text-editor
""")

    # 4) update desktop database (some distros)
    subprocess.run(["update-desktop-database", str(app_dir)], check=True)

    gw.info(".gwr association registered in your Linux desktop environment.")





import unittest
import jwt
from gway import gw

web = gw.load_project('web')

class JwtAuthChallengeTests(unittest.TestCase):
    class FakeRequest:
        def __init__(self, token):
            self.headers = {'authorization': f'Bearer {token}'}

    class FakeResponse:
        def __init__(self):
            self.status_code = None
            self.headers = {}

    def test_valid_token(self):
        secret = 'test-secret'
        token = jwt.encode({'sub': 'alice'}, secret, algorithm='HS256')
        web.auth.clear()
        web.auth.config_jwt(secret=secret, algorithms=['HS256'], engine='fastapi')
        ok = web.auth.is_authorized(strict=True, context={'request': self.FakeRequest(token), 'response': self.FakeResponse()})
        self.assertTrue(ok)
        web.auth.clear()

    def test_invalid_token(self):
        secret = 'test-secret'
        token = jwt.encode({'sub': 'alice'}, 'wrong', algorithm='HS256')
        web.auth.clear()
        web.auth.config_jwt(secret=secret, algorithms=['HS256'], engine='fastapi')
        ok = web.auth.is_authorized(strict=True, context={'request': self.FakeRequest(token), 'response': self.FakeResponse()})
        self.assertFalse(ok)
        web.auth.clear()

if __name__ == '__main__':
    unittest.main()



import unittest
import json
from gway import gw
import unittest.mock as mock
import importlib.util


class QPigFarmButtonTests(unittest.TestCase):
    def setUp(self):
        self.time_patch = mock.patch('time.time', lambda: 0)
        self.time_patch.start()
        # Load the underlying module for constants
        path = gw.resource('projects', 'games', 'qpig.py')
        spec = importlib.util.spec_from_file_location('qpig_mod', str(path))
        self.qpig_mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.qpig_mod)

    def tearDown(self):
        self.time_patch.stop()

    def test_buy_small_and_adopt(self):
        state = {
            'pigs': self.qpig_mod.DEFAULT_PIGS,
            'mc': 1000.0,
            'pellets': 0.0,
            'time': 0.0,
            'avail': self.qpig_mod.DEFAULT_AVAILABLE,
            'enc_small': self.qpig_mod.DEFAULT_ENC_SMALL,
            'enc_large': self.qpig_mod.DEFAULT_ENC_LARGE,
            'last_add': 0.0,
            'veggies': {},
            'food': [],
            'offer': {'kind': 'carrot', 'qty': 1, 'price': 10, 'time': 0.0},
        }
        state = self.qpig_mod._process_state(state, 'buy_small')
        self.assertEqual(state['enc_small'], 2)
        self.assertEqual(state['mc'], 1000 - self.qpig_mod.SMALL_COST)

        state = self.qpig_mod._process_state(state, 'adopt')
        self.assertEqual(state['pigs'], 2)
        self.assertEqual(state['avail'], self.qpig_mod.DEFAULT_AVAILABLE - 1)

    def test_buy_and_place_veggie(self):
        state = {
            'pigs': self.qpig_mod.DEFAULT_PIGS,
            'mc': 1000.0,
            'pellets': 0.0,
            'time': 0.0,
            'avail': self.qpig_mod.DEFAULT_AVAILABLE,
            'enc_small': self.qpig_mod.DEFAULT_ENC_SMALL,
            'enc_large': self.qpig_mod.DEFAULT_ENC_LARGE,
            'last_add': 0.0,
            'veggies': {},
            'food': [],
            'offer': {'kind': 'carrot', 'qty': 2, 'price': 10, 'time': 0.0},
        }
        state = self.qpig_mod._process_state(state, 'buy_veggie')
        self.assertEqual(state['veggies'].get('carrot'), 2)
        self.assertEqual(state['mc'], 1000 - 20)

        state = self.qpig_mod._process_state(state, 'place_carrot')
        self.assertEqual(state['veggies'].get('carrot'), 1)
        self.assertEqual(len(state['food']), 1)
        self.assertEqual(state['food'][0][0], 'carrot')

    def test_view_contains_canvas(self):
        html = self.qpig_mod.view_qpig_farm()
        self.assertIn("qpig-canvas", html)

if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch
from gway import gw

class StaticCollectTests(unittest.TestCase):
    def test_collect_concatenates_files(self):
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            static_root = tmp_path / "data" / "static"
            proj_dir = static_root / "web" / "site"
            proj_dir.mkdir(parents=True)
            css1 = proj_dir / "a.css"
            css2 = proj_dir / "b.css"
            css1.write_text("body{color:red}")
            css2.write_text("p{font-weight:bold}")
            js1 = proj_dir / "a.js"
            js2 = proj_dir / "b.js"
            js1.write_text("console.log('a');")
            js2.write_text("console.log('b');")
            target_dir = tmp_path / "work" / "shared"
            target_dir.mkdir(parents=True)

            def fake_resource(*parts, **kw):
                return tmp_path.joinpath(*parts)

            with patch.object(gw, "resource", fake_resource), \
                 patch.object(gw.web.app, "enabled_projects", lambda: {"web.site"}):
                report = gw.web.static.collect(root="data/static", target="work/shared")

            self.assertEqual(
                {Path(rel).as_posix() for _, rel, _ in report["css"]},
                {"web/site/a.css", "web/site/b.css"},
            )
            self.assertEqual(
                {Path(rel).as_posix() for _, rel, _ in report["js"]},
                {"web/site/a.js", "web/site/b.js"},
            )

            css_bundle = Path(report["css_bundle"]).read_text()
            expected_css = "".join(
                f"/* --- {proj}:{rel} --- */\n" + Path(full).read_text() + "\n\n"
                for proj, rel, full in reversed(report["css"])
            )
            self.assertEqual(css_bundle, expected_css)

            js_bundle = Path(report["js_bundle"]).read_text()
            expected_js = "".join(
                f"// --- {proj}:{rel} ---\n" + Path(full).read_text() + "\n\n"
                for proj, rel, full in report["js"]
            )
            self.assertEqual(js_bundle, expected_js)

    def test_collect_includes_monitor_tabs_script(self):
        """net_monitors.js is bundled when monitor project is enabled."""
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            monitor_dir = tmp_path / "data" / "static" / "monitor"
            monitor_dir.mkdir(parents=True)
            js_file = monitor_dir / "net_monitors.js"
            js_file.write_text("console.log('tabs');")
            css_file = monitor_dir / "net_monitors.css"
            css_file.write_text(".tabs{}")
            target_dir = tmp_path / "work" / "shared"
            target_dir.mkdir(parents=True)

            def fake_resource(*parts, **kw):
                return tmp_path.joinpath(*parts)

            with patch.object(gw, "resource", fake_resource), \
                 patch.object(gw.web.app, "enabled_projects", lambda: {"monitor"}):
                report = gw.web.static.collect(root="data/static", target="work/shared")

            js_files = {Path(rel).as_posix() for _, rel, _ in report["js"]}
            self.assertIn("monitor/net_monitors.js", js_files)
            js_bundle = Path(report["js_bundle"]).read_text()
            self.assertIn("net_monitors.js", js_bundle)

if __name__ == "__main__":
    unittest.main()



import unittest
import importlib.util
from pathlib import Path
import sys
import types


class CDVUtilsTests(unittest.TestCase):
    @staticmethod
    def _load_cdv():
        site_spec = importlib.util.spec_from_file_location(
            "projects.web.site",
            Path(__file__).resolve().parents[1] / "projects" / "web" / "site.py",
        )
        site_mod = importlib.util.module_from_spec(site_spec)
        site_spec.loader.exec_module(site_mod)

        projects_mod = types.ModuleType("projects")
        web_mod = types.ModuleType("projects.web")
        web_mod.site = site_mod
        projects_mod.web = web_mod
        sys.modules.setdefault("projects", projects_mod)
        sys.modules.setdefault("projects.web", web_mod)
        sys.modules.setdefault("projects.web.site", site_mod)

        cdv_path = Path(__file__).resolve().parents[1] / "projects" / "cdv.py"
        spec = importlib.util.spec_from_file_location("cdv_mod", cdv_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
    @classmethod
    def setUpClass(cls):
        cls.cdv_mod = cls._load_cdv()
    def test_sanitize_cdv_path(self):
        cases = {
            "foo/bar": "foo/bar",
            "../secret/file.cdv": "secret/file.cdv",
            "foo\\bar": "foo/bar",
            "../../evil": "evil",
        }
        for raw, expected in cases.items():
            with self.subTest(raw=raw):
                self.assertEqual(self.cdv_mod._sanitize_cdv_path(raw), expected)

    def test_parse_and_serialize_roundtrip(self):
        text = "id1:name=John%20Doe:age=30\nid2:msg=Hello"
        records = self.cdv_mod._parse_cdv_text(text)
        self.assertEqual(records, {
            "id1": {"name": "John Doe", "age": "30"},
            "id2": {"msg": "Hello"},
        })
        serialized = self.cdv_mod._records_to_text(records)
        self.assertEqual(serialized, "id1:name=John%20Doe:age=30\nid2:msg=Hello")


if __name__ == "__main__":
    unittest.main()



"""Utility helpers for automated website testing using Selenium."""

from __future__ import annotations

from contextlib import contextmanager
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as FirefoxService
from webdriver_manager.firefox import GeckoDriverManager

DEFAULT_BROWSER = "chrome"


def get_driver(*, browser: str = DEFAULT_BROWSER, headless: bool = True):
    """Return a Selenium WebDriver instance for the selected browser."""
    browser = browser.lower()
    if browser == "chrome":
        options = ChromeOptions()
        if headless:
            options.add_argument("--headless")
        options.add_argument("--no-sandbox")
        options.add_argument("--disable-dev-shm-usage")
        service = ChromeService(ChromeDriverManager().install())
        return webdriver.Chrome(service=service, options=options)
    elif browser == "firefox":
        options = FirefoxOptions()
        if headless:
            options.add_argument("--headless")
        service = FirefoxService(GeckoDriverManager().install())
        return webdriver.Firefox(service=service, options=options)
    else:
        raise ValueError(f"Unsupported browser: {browser}")


_active_driver = None
_active_config: tuple[str, bool] | None = None


@contextmanager
def browse(
    *,
    url: str | None = None,
    browser: str = DEFAULT_BROWSER,
    headless: bool = True,
    driver=None,
    close: bool = False,
):
    """Yield a cached WebDriver instance, creating a new one if needed."""
    global _active_driver, _active_config

    # Try to unwrap an existing driver from the given object (tuple/list etc)
    try:
        match driver:
            case webdriver.Remote() as d:
                driver = d
            case list() | tuple() as seq:
                driver = next((x for x in seq if isinstance(x, webdriver.Remote)), None)
            case None:
                driver = None
            case _ if isinstance(driver, webdriver.Remote):
                pass
            case _ if hasattr(driver, "__iter__") and not isinstance(driver, (str, bytes, bytearray)):
                driver = next((x for x in driver if isinstance(x, webdriver.Remote)), None)
            case _:
                driver = None
    except Exception:
        pass

    if close:
        target = driver or _active_driver
        if target:
            try:
                target.quit()
            except Exception:
                pass
            if target is _active_driver:
                _active_driver = None
                _active_config = None
        yield None
        return

    desired_cfg = (browser.lower(), bool(headless))

    if driver:
        _active_driver = driver
        _active_config = desired_cfg
    else:
        if (
            _active_driver is None
            or _active_config != desired_cfg
            or getattr(_active_driver, "service", None) is None
        ):
            if _active_driver:
                try:
                    _active_driver.quit()
                except Exception:
                    pass
            _active_driver = get_driver(browser=browser, headless=headless)
            _active_config = desired_cfg

    if url:
        _active_driver.get(url)

    try:
        yield _active_driver
    finally:
        # Do not quit driver here; it can be reused. Call with driver=None
        # and contradictory params to force recreation.
        pass


def capture_page_source(
    url: str,
    *,
    browser_name: str = DEFAULT_BROWSER,
    headless: bool = True,
    wait: float = 2.0,
    screenshot: str | None = None,
) -> str:
    """Return page source for ``url``. Optionally save a screenshot."""
    import time

    with browse(url=url, browser=browser_name, headless=headless) as drv:
        if wait:
            time.sleep(wait)
        if screenshot:
            drv.save_screenshot(screenshot)
        return drv.page_source



import unittest
import os
import sys
from io import StringIO
from unittest.mock import patch
from gway import gw

class NotifyBuiltinTests(unittest.TestCase):
    def setUp(self):
        self.out = StringIO()
        self.orig = sys.stdout
        sys.stdout = self.out

    def tearDown(self):
        sys.stdout = self.orig
        os.environ.pop('ADMIN_EMAIL', None)

    def test_console_fallback(self):
        with patch.object(gw.screen, 'notify', side_effect=Exception('fail')):
            with patch.object(gw.mail, 'send') as mock_send:
                gw.notify('hello world', title='T')
                mock_send.assert_not_called()
        self.assertIn('hello world', self.out.getvalue())

    def test_email_fallback(self):
        os.environ['ADMIN_EMAIL'] = 'test@example.com'
        with patch.object(gw.screen, 'notify', side_effect=Exception('fail')):
            with patch.object(gw.mail, 'send') as mock_send:
                gw.notify('msg', title='Notice')
                mock_send.assert_called_once()

if __name__ == '__main__':
    unittest.main()



# file: projects/web/message.py

"""Cookie-backed message banner utilities."""

import html
import re
from gway import gw


def write(text: str, *, sep: str = "\n", max_messages: int = 4, max_len: int = 180) -> None:
    """Store a short message in the ``message`` cookie if the project is enabled."""
    app = getattr(gw.web, "app", None)
    if not (app and getattr(app, "is_setup", lambda x: False)("web.message")):
        gw.warning("web.message not enabled; message discarded")
        return
    if not (app.is_setup("web.cookies") and gw.web.cookies.accepted()):
        gw.warning(f"Cannot store message without cookie consent: {text!r}")
        return

    if not isinstance(text, str):
        text = str(text)
    clean = re.sub(r"\s+", " ", text.strip())
    clean = re.sub(r"[<>]", "", clean)
    if len(clean) > max_len:
        clean = clean[:max_len]
    if not clean:
        return

    raw = gw.web.cookies.get("message", "") or ""
    msgs = [m for m in raw.split(sep) if m]
    msgs.append(clean)
    if len(msgs) > max_messages:
        msgs = msgs[-max_messages:]
    gw.web.cookies.set("message", sep.join(msgs))


def render(*, sep: str = "\n") -> str:
    """Return HTML for any stored messages, or an empty string."""
    app = getattr(gw.web, "app", None)
    if not (app and getattr(app, "is_setup", lambda x: False)("web.message")):
        return ""
    if not app.is_setup("web.cookies"):
        return ""
    try:
        raw_msg = gw.web.cookies.get("message", "")
    except Exception:
        raw_msg = ""
    if not raw_msg:
        return ""
    parts = [html.escape(m.strip()) for m in raw_msg.split(sep) if m.strip()]
    if not parts:
        return ""
    rows = "".join(f"<div>{p}</div>" for p in parts)
    return (
        "<div id='gw-message' style='background:#fffae6;color:#000;"
        "border:2px solid #d69e00;padding:0.7em 1.6em 0.7em 1em;"
        "margin-bottom:1em;position:relative;font-weight:bold;'>"
        + rows +
        "<a href='#' onclick=\"gwDismissMessage();return false;\" "
        "style='position:absolute;top:0.2em;right:0.6em;color:#900;"
        "text-decoration:none;font-size:1.2em;font-weight:bold;'>\u2715</a>"
        "</div>"
        "<script>function gwDismissMessage(){document.cookie='message=;path=/;"
        "expires=Thu, 01 Jan 1970 00:00:00 GMT';var m=document.getElementById('gw-message');"
        "if(m)m.style.display='none';}</script>"
    )



import unittest
import sys
from gway import gw
from paste.fixture import TestApp

class SetupHomeLinksFuncTests(unittest.TestCase):
    def test_defaults_from_project_functions(self):
        app = gw.web.app.setup_app("dummy")
        mod = sys.modules[gw.web.app.setup_app.__module__]
        self.assertIn(("Dummy", "dummy/index"), mod._homes)
        self.assertEqual(mod._links.get("dummy/index"), ["about", "more"])
        client = TestApp(app)
        resp = client.get("/dummy")
        self.assertEqual(resp.status, 200)
        self.assertIn("Dummy Index", resp.body.decode())

if __name__ == "__main__":
    unittest.main()
